/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2017 Wolfgang Betz
 *
 * Fesslix is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fesslix is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fesslix.  If not, see <http://www.gnu.org/licenses/>. 
 */


#define flxbn_flxbn_manager_CPP

#include "flxbn_manager.h"
#include "flxmath_rnd.h"

#include <limits>

// using namespace std;
using namespace flxbn;

tuint* bn_rv_base::stateIDs = NULL;


bn_node::~bn_node()
{
  if (next) delete next;
}

void bn_node::check_cpt(flxVec& cpt, const tuint nrows, const tuint ncols, const tdouble roundoff_correct) const
{
  // check table size
    if (cpt.get_N()!=ncols*nrows) {
      std::ostringstream ssV;
      ssV << cpt.get_N() << " vs. " << nrows << "*" << ncols << "=" << nrows*ncols;
      throw FlxException("bn_node::check_cpt_1", "Size of tables does not match.", ssV.str() );
    }
  for (tuint j=0;j<ncols;++j) {
    tdouble sum=ZERO;
    for (tuint i=0;i<nrows;++i) {
      const tdouble t = cpt[i*ncols+j];
      if (t<ZERO) {
	std::ostringstream ssV;
	ssV << "\tcol=" << j+1 << " (starts counting from 1)" << std::endl;
	ssV << "\trow=" << i+1 << " (starts counting from 1)" << std::endl;
	ssV << "\tval=" << GlobalVar.Double2String(t);
	throw FlxException("bn_node::check_cpt_2","Invalid probability table (negative value).",ssV.str());
      }
      sum+=t;
    }
    // round-off correction
      if (fabs(sum-ONE)<roundoff_correct) {
	const tdouble d = (ONE-sum)/sum+ONE;
	  std::ostringstream ssV;
	  ssV << "Corrected column in probability table:\n";
	  ssV << "  col=" << j+1 << " (starts counting from 1)\n";
	  ssV << "  factor=" << GlobalVar.Double2String(d-ONE) << std::endl;
	  GlobalVar.alert.alert("bn_node::check_cpt_3",ssV.str());
	sum = ZERO;
	for (tuint i=0;i<nrows;++i) {
	  tdouble& t = cpt[i*ncols+j];
	  t *= d;
	  sum += t;
	}
      }
    // is there an error?
      if (fabs(sum-ONE)>GlobalVar.TOL()) {
	std::ostringstream ssV;
	ssV << "\tcol=" << j+1 << " (starts counting from 1)" << std::endl;
	ssV << "\tsum-1=" << GlobalVar.Double2String(sum-ONE);
	throw FlxException("bn_node::check_cpt_4","Invalid probability table.",ssV.str());
      }
    }
}

rv_edge& rv_edge::operator=(const flxbn::rv_edge& rhs)
{
  throw FlxException_Crude("rv_edge::operator=");
}

void rv_edge::clear_topo(const bool full, const flxbn::bn_rv_base*const owner)
{
  const int pown = owner->get_streeID_new();
  const int prv = rv->get_streeID_new();
  #if FLX_DEBUG
    if (!(pown==0||pown==-1) || !(prv==0||prv==-1)) throw FlxException_Crude("rv_edge::clear_topo");
  #endif
  if (pown==-1&&prv==-1) {
    if (full) btreeID = -1;
    streeID = -1;
  } else {
    if (full) btreeID = 0;
    streeID = 0;
  }
}


void rv_obs_pair::observe() const
{
  bnm.observe_RV(rv,observedStates);
}


bn_rv_base::bn_rv_base(const std::string name, const bool outputRV, const std::vector< std::string >& states, const std::vector< bn_rv_base* > drvs, bn_node*const node)
: ID(0), btreeID(0), streeID(0), parentN(drvs.size()), childN(0), obsrvType(noObsrv), locked(false), name(name), outputRV(outputRV), stateN(states.size()), states(states), node(node) 
{
  if (stateIDs) {
    delete [] stateIDs;
    stateIDs = NULL;
  }
  bnm.topo_reset();
  // set-up parents
    for (tuint i=0;i<parentN;++i) {
      parents.push_back(rv_edge(drvs[i]));
      drvs[i]->set_child(this);
    }
}

const bool bn_rv_base::is_state_observed(const tuint stateID) const
{
  const size_t S = observedStates.size();
  for (tuint i = 0; i < S; ++i) {
    if (observedStates[i]==stateID) return true;
  }
  return false;
}

void bn_rv_base::set_child(bn_rv_base* crv)
{
  childN++;
  childs.push_back(rv_edge(crv));
  bnm.topo_reset();
}

rv_edge& bn_rv_base::get_child(bn_rv_base* crv)
{
  for (tuint i=0;i<childN;++i) {
    if (childs[i].rv==crv) return childs[i];
  }
  throw FlxException_Crude("bn_rv_base::get_child");
}

rv_edge& bn_rv_base::get_parent(bn_rv_base* prv)
{
  for (tuint i=0;i<parentN;++i) {
    if (parents[i].rv==prv) return parents[i];
  }
  throw FlxException_Crude("bn_rv_base::get_parent");
}

void bn_rv_base::set_ID(const tuint IDv)
{
  ID = IDv; 
  bnm.topo_reset();
}

void bn_rv_base::observe(const tuint sid)
{
  std::vector< tuint > oV;	// observation
  oV.push_back(sid);
  observe(oV);
}

void bn_rv_base::observe(const std::vector< tuint > oV)
{
  if (oV.empty()) {
    unobserve();
    return;
  }
  if (oV.size()==1) {
    obsrvType = snglObsrv;
    if (observedStates.size()!=1) {
      observedStates.resize(1);
    }
    observedStates[0] = oV[0];
    if (stateIDs) stateIDs[ID] = oV[0];
    if (get_type()==decNode) return;
  } else {
    if (oV.size()==stateN) {
      GlobalVar.alert.alert("bn_rv_base::observe","All states observed ... this is equal to unobserved!");
      unobserve();
      return;
    }
    obsrvType = multObsrv;
    observedStates.resize(oV.size());
    for (tuint i=0;i<oV.size();++i) {
      observedStates[i] = oV[i];
    }
    std::sort(observedStates.begin(), observedStates.end());  
  }
  bnm.topo_reset();
}

void bn_rv_base::unobserve()
{
  if (obsrvType==noObsrv) return;
  bnm.topo_reset();
  obsrvType = noObsrv;
  observedStates.empty();
}

void bn_rv_base::clear_topo(const bool full)
{
  if (full) bnm.topo_reset();
  if (full) btreeID = 0;
  streeID = 0;
  for (tuint i=0;i<childN;++i) {
    childs[i].clear_topo(full,this);
  }
  for (tuint i=0;i<parentN;++i) {
    parents[i].clear_topo(full,this);
  }
}

void bn_rv_base::stree_remove()
{
  for (tuint i=0;i<parentN;++i) {
    parents[i].streeID = -2;
    parents[i].rv->get_child(this).streeID = -2;
  }
  for (tuint i=0;i<childN;++i) {
    childs[i].streeID = -2;
    childs[i].rv->get_parent(this).streeID = -2;
  }
  streeID = -2;
}

void bn_rv_base::detect_topo(const int tid, const int tdir, bn_rv_base* last)
{
  #if FLX_DEBUG
    if (tid<=0) throw FlxException_Crude("bn_rv_base::detect_topo_0");
  #endif
  bnm.topo_reset();
  // update corresponding (last) child or parent
    if (tdir==1) {		// downwards (check parents)
      #if FLX_DEBUG
	bool b=false;
      #endif
      for (tuint i=0;i<parentN;++i) {
	if (parents[i].rv == last) {
	  #if FLX_DEBUG
	    b=true;
	  #endif
	  parents[i].btreeID = tid;
	  break;
	}
      }
      #if FLX_DEBUG
	if (!b) {
	  throw FlxException_Crude("bn_rv_base::detect_topo_1");
	}
      #endif
    } else if (tdir==2) {	// upwards (check childs)
      #if FLX_DEBUG
	bool b=false;
      #endif
      for (tuint i=0;i<childN;++i) {
	if (childs[i].rv == last) {
	  #if FLX_DEBUG
	    b=true;
	  #endif
	  childs[i].btreeID = tid;
	  break;
	}
      }
      #if FLX_DEBUG
	if (!b) {
	  throw FlxException_Crude("bn_rv_base::detect_topo_2");
	}
      #endif
    }
  // set btreeID 
    if (obsrvType==snglObsrv && tdir==2) return;
    if (btreeID==0) {		// not in tree, so far
	btreeID = tid;
    } else if (btreeID==tid) {			// there is a circle !!!
      return;
    } else {
      throw FlxException_Crude("bn_rv_base::detect_topo_3");
    }
  // add further RVs
    // explaining away
      if (obsrvType==snglObsrv) {
	if (tdir!=2 && parentN>1) {		// consider other parents as well
	  for (tuint i=0;i<parentN;++i) {
	    if (parents[i].btreeID == 0) {		
	      parents[i].btreeID = tid;
	      parents[i].rv->detect_topo(tid,2,this);
	    } else if (parents[i].btreeID == tid) {	// do not go back by accident
	    } else if (parents[i].btreeID < 0) {
	    } else {
	      throw FlxException_Crude("bn_rv_base::detect_topo_4");
	    }
	  }
	}
	return;
      }
    // add parents
      for (tuint i=0;i<parentN;++i) {
	if (parents[i].btreeID == 0) {		
	  parents[i].btreeID = tid;
	  parents[i].rv->detect_topo(tid,2,this);
	} else if (parents[i].btreeID == tid) {	// do not go back by accident
	} else if (parents[i].btreeID < 0) {
	  throw FlxException_Crude("bn_rv_base::detect_topo_5");
	} else {
	  throw FlxException_Crude("bn_rv_base::detect_topo_6");
	}
      }
    // add childs
      for (tuint i=0;i<childN;++i) {
	if (childs[i].btreeID == 0) {		
	  childs[i].btreeID = tid;
	  childs[i].rv->detect_topo(tid,1,this);
	} else if (childs[i].btreeID == tid) {	// do not go back by accident
	} else if (childs[i].btreeID < 0) {
	  throw FlxException_Crude("bn_rv_base::detect_topo_7");
	} else {
	  throw FlxException_Crude("bn_rv_base::detect_topo_8");
	}
      }
}

void bn_rv_base::detect_topo_sub(const int btid, const int stid, const int tdir, bn_rv_base* last)
{
  #if FLX_DEBUG
    if (btid<=0 || stid<=0) throw FlxException_Crude("bn_rv_base::detect_topo_sub_0");
  #endif
  bnm.topo_reset();
  // update corresponding (last) child or parent
    if (tdir==1) {		// downwards (check parents)
      #if FLX_DEBUG
	bool b=false;
      #endif
      for (tuint i=0;i<parentN;++i) {
	if (parents[i].rv == last) {
	  #if FLX_DEBUG
	    if (parents[i].btreeID != btid) throw FlxException_Crude("bn_rv_base::detect_topo_sub_1");
	    b=true;
	  #endif
	  parents[i].streeID = stid;
	  break;
	}
      }
      #if FLX_DEBUG
	if (!b) {
	  throw FlxException_Crude("bn_rv_base::detect_topo_sub_2");
	}
      #endif
    } else if (tdir==2) {	// upwards (check childs)
      #if FLX_DEBUG
	bool b=false;
      #endif
      for (tuint i=0;i<childN;++i) {
	if (childs[i].rv == last) {
	  #if FLX_DEBUG
	    if (childs[i].btreeID != btid) throw FlxException_Crude("bn_rv_base::detect_topo_sub_3");
	    b=true;
	  #endif
	  childs[i].streeID = stid;
	  break;
	}
      }
      #if FLX_DEBUG
	if (!b) {
	  throw FlxException_Crude("bn_rv_base::detect_topo_sub_4");
	}
      #endif
    }
  // set streeID 
    if (streeID==0) {		// not in tree, so far
      if (obsrvType==snglObsrv && tdir==2) return;
      streeID = stid;
    } else if (streeID==stid) {			// there is a circle !!!
      return;
    } else {
      throw FlxException_Crude("bn_rv_base::detect_topo_sub_7");
    }
    #if FLX_DEBUG
      if (btreeID!=btid) throw FlxException_Crude("bn_rv_base::detect_topo_sub_5");
    #endif
  // add further RVs
    // explaining away
      if (obsrvType==snglObsrv) {
	if (tdir!=2 && parentN>1) {		// consider other parents as well
	  for (tuint i=0;i<parentN;++i) {
	    if (parents[i].streeID == 0) {		
	      parents[i].streeID = stid;
	      parents[i].rv->detect_topo_sub(btid,stid,2,this);
	    } else if (parents[i].streeID == stid) {	// do not go back by accident
	    } else if (parents[i].streeID < 0) {
	    } else {
	      throw FlxException_Crude("bn_rv_base::detect_topo_sub_8");
	    }
	  }
	}
	return;
      }
    // add parents
      for (tuint i=0;i<parentN;++i) {
	if (parents[i].streeID == 0) {		
	  parents[i].streeID = stid;
	  parents[i].rv->detect_topo_sub(btid,stid,2,this);
	} else if (parents[i].streeID == stid) {	// do not go back by accident
	} else if (parents[i].streeID < 0) {
	  throw FlxException_Crude("bn_rv_base::detect_topo_sub_9");
	} else {
	  throw FlxException_Crude("bn_rv_base::detect_topo_sub_10");
	}
      }
    // add childs
      for (tuint i=0;i<childN;++i) {
	if (childs[i].streeID == 0) {		
	  childs[i].streeID = stid;
	  childs[i].rv->detect_topo_sub(btid,stid,1,this);
	} else if (childs[i].streeID == stid) {	// do not go back by accident
	} else if (childs[i].streeID == -2) {
	} else if (childs[i].streeID == -1) {
	  throw FlxException_Crude("bn_rv_base::detect_topo_sub_11");
	} else {
	  throw FlxException_Crude("bn_rv_base::detect_topo_sub_12");
	}
      }
}

const tuint bn_rv_base::get_dist(const bn_rv_base*const drv) const
{
  if (drv==this) return 0;
  // search in parents
    for (tuint i=0;i<parentN;++i) {
      if (parents[i].rv == drv) {
	if ((drv->get_type()==trueRV&&obsrvType==snglObsrv)||drv->get_type()!=trueRV) {
	  return 0;
	} else {
	  return 1;
	}
      }
    }
  // search in childs (and spouses)
    for (tuint i=0;i<childN;++i) {
      if (childs[i].rv == drv) {	// childs
	if ((drv->get_type()==trueRV&&obsrvType==snglObsrv)||drv->get_type()!=trueRV) {
	  return 0;
	} else {
	  return 1;
	}
      } else {				// spouses
	bn_rv_base* crv = childs[i].rv;
	for (tuint j=0;j<crv->parentN;++j) {
	  if (crv->parents[j].rv == drv) {
	    if ((drv->get_type()==trueRV&&obsrvType==snglObsrv)||drv->get_type()!=trueRV) {
	      return 0;
	    } else {
	      return 1;
	    }
	  }
	}
      }
    }
  // return infinity
    return std::numeric_limits<tuint>::max();
}

void flxbn::bn_rv_base::get_subtree_rvs(bool* vist, const int streeIDv) const
{
  vist[ID] = true;
  // check parents
    for (tuint i=0;i<parentN;++i) {
      if (parents[i].rv->streeID==streeIDv) {
	if (parents[i].rv->obsrvType!=snglObsrv && !vist[parents[i].rv->ID]) {
	  parents[i].rv->get_subtree_rvs(vist,streeIDv);
	}
      }
    }
  if (streeID!=-2) {
    if (obsrvType==snglObsrv) return;
    // check childs
      for (tuint i=0;i<childN;++i) {
	if ( (childs[i].streeID==streeIDv || childs[i].streeID==-2) && !vist[childs[i].rv->ID]) {
	  childs[i].rv->get_subtree_rvs(vist,streeIDv);
	}
      }
  }
}

const tuint bn_rv_base::get_stateID(const std::string ts) const
{
  for (tuint i=0;i<stateN;++i) {
    if (states[i]==ts) {
      return i;
    }
  }
  std::ostringstream ssV;
  ssV << "Unknown state '" << ts << "'.";
  throw FlxException("bn_rv_base::get_stateID", ssV.str() );
}

const int bn_rv_base::get_streeID_new() const
{
  if (obsrvType==snglObsrv) return -1;
  else return 0;
}



bn_rv::bn_rv(const std::string name, const bool outputRV, const std::vector< std::string >& states, const std::vector< bn_rv_base* > drvs, const flxVec& tableV, bn_node*const node)
: bn_rv_base(name,outputRV,states,drvs,node), tableCols(1), tableColsVec(drvs.size()),tableCol0(tableV.get_N()/states.size()), tableCol0sum(tableCol0.size()), table(tableV) 
{
  // set-up parents
    const tuint pN = get_parentN();
    for (tuint i=0;i<pN;++i) {
      tableColsVec[pN-1-i] = tableCols;
      tableCols*=drvs[pN-1-i]->stateN;
    }
    if (tableCols!=tableCol0.size()) {
      std::ostringstream ssV;
      ssV << "name=" << name << std::endl;
      ssV << "\tsize=" << tableCols << " -> expected: " << tableCol0.size() << "=" << tableV.get_N() << "/" << states.size();
      throw FlxException("bn_rv::bn_rv_1", "Size of probability table incorrect.", ssV.str() );
    }
  update_tableCol0();
}

inline const tuint bn_rv::get_curr_col() const
{
  const tuint pN = get_parentN();
  if (pN==0) {
    return 0;
  } else {
    tuint col = 0;
    for (tuint i=0;i<pN;++i) {
      col += tableColsVec[i]*stateIDs[get_parent(i).rv->get_ID()];
    }
    return col;
  }
}

void flxbn::bn_rv::update_tableCol0()
{
  if (get_obsrvType()!=multObsrv ) {
    for (tuint j=0;j<tableCols;++j) {
      tuint c=0;
      for (tuint i=0;i<stateN;++i) {
	
	if (table[i*tableCols+j]>0) ++c;
      }
      tableCol0[j]=c;
    }
  } else {
    for (tuint j=0;j<tableCols;++j) {
      tuint a=0; 	// last observed variable
      const tuint aN = get_observedStatesN()-1;
      tuint c=0;
      tdouble s = ZERO;
      for (tuint i=0;i<stateN;++i) {
	if (a<aN && get_observedStates(a)<i) a++;
	if (get_observedStates(a)==i) {
	  const tdouble t = table[i*tableCols+j];
	  if (t>0) {
	    ++c;
	    s+=t;
	  }
	}
      }
      tableCol0[j] = c;
      tableCol0sum[j] = s;
    }
  }
}

void bn_rv::observe(const std::vector< tuint > oV)
{
  flxbn::bn_rv_base::observe(oV);
  update_tableCol0();
}

void bn_rv::unobserve()
{
  flxbn::bn_rv_base::unobserve();
  update_tableCol0();
}

const bool bn_rv::rnd_state(tdouble& p) const
{
  const tuint col = get_curr_col();
  switch (get_obsrvType()) {
    case noObsrv:
    {
      const tdouble t = rv_uniform();
      tdouble c = ZERO;
      tuint i;
      for (i=0;i<stateN-1;++i) {
	c += table[tableCols*i+col];
	if (c>0 && c>=t) break;
      }
      stateIDs[get_ID()] = i;
      return true;
    }
    case snglObsrv:
      p*=table[tableCols*stateIDs[get_ID()]+col];
      return (p>0);
    case multObsrv :
    {
      const tdouble t = rv_uniform();
      tdouble c = ZERO;
      tuint a=0;	// last observed variable
      const tuint aN = get_observedStatesN()-1;
      tuint i; 
      for (i=0;i<stateN-1;++i) {
	if (a<aN && get_observedStates(a)<i) a++;
	c += table[tableCols*i+col];
	if (c>0 && c>=t) break;
      }
      if (a<aN && get_observedStates(a)<i) a++;
      if (get_observedStates(a)==i) {
	stateIDs[get_ID()] = i;
	return true;
      } else {
	p=ZERO;
	return false;
      }
    }
    default:
      #if FLX_DEBUG
	throw FlxException_Crude("bn_rv::rnd_state_2");
      #else
	return true;
      #endif
  }
}

void bn_rv::rnd_state(tdouble& p, tdouble& p2) const
{
  const tuint col = get_curr_col();
  switch (get_obsrvType()) {
    case noObsrv:
    {
      while (true) {
	const tdouble t = rv_uniform();
	tuint cS = tuint(t*stateN);
	if (cS>=stateN) cS = stateN-1;
	const tdouble tp = table[tableCols*cS+col];
	if (tp==ZERO) continue;
	const tdouble mod = tp*tableCol0[col];
	p*=mod;
	p2*=mod;
	stateIDs[get_ID()] = cS;
	return;
      }
    }
    case snglObsrv:
      p*=table[tableCols*stateIDs[get_ID()]+col];
      return;
    case multObsrv :
    {
      if (tableCol0[col]==0) {
	p=ZERO;
	return;
      }
      const tuint aN = get_observedStatesN();
      while (true) {
	const tdouble t = rv_uniform();
	tuint cS = tuint(t*stateN);
	if (cS>=stateN) cS = stateN-1;
	// check if state is observed
	  bool f=false;
	  for (tuint i=0;i<aN;++i) {
	    const tuint oS = get_observedStates(i);
	    if (oS==cS) {
	      f=true;
	      break;
	    } else if (oS>cS) {
	      break;
	    }
	  }
	  if (!f) continue;
	const tdouble tp = table[tableCols*cS+col];
	if (tp==ZERO) continue;
	const tdouble mod = tp*tableCol0[col];
	p*=mod;
	p2*=mod/tableCol0sum[col];
	stateIDs[get_ID()] = cS;
	return;
      }
    }
    default:
      #if FLX_DEBUG
	throw FlxException_Crude("bn_rv::rnd_state_2");
      #else	
	return;
      #endif
  }
}



bn_detRel::bn_detRel(const std::string name, const bool outputRV, const std::vector< std::string >& states, const std::vector< bn_rv_base* > drvs, bn_node*const node)
: bn_rv_base(name,outputRV,states,drvs,node)
{

}

const bool bn_detRel::depends_on(bn_rv_base* rv)
{
  if (rv==this) return true;
  for (tuint i=0;i<stateN;++i) {
    if (get_parent(i).rv == rv) return true;
  }
  return false;
}

bn_detRel_and::bn_detRel_and(const std::string name, const bool outputRV, const std::vector< std::string >& states, const std::vector< bn_rv_base* > drvs, bn_node*const node)
: bn_detRel(name, outputRV, states, drvs, node)
{
  // check bool-property
    if (states.size()!=2) {
      std::ostringstream ssV;
      ssV << "Number of states must be equal to two.";
      throw FlxException("bn_detRel_and::bn_detRel_and_1", ssV.str() );
    }
    for (tuint i=0;i<drvs.size();++i) {
      if (drvs[i]->stateN!=2) {
	std::ostringstream ssV;
	ssV << "Number of states of parent variable '" << drvs[i]->name << "' must be equal to two.";
	throw FlxException("bn_detRel_and::bn_detRel_and_2", ssV.str() );
      }
    }
}

const bool bn_detRel_and::eval() const
{
  for (tuint i=0;i<get_parentN();++i) {
    if (stateIDs[get_parent(i).rv->get_ID()]==0) return false;
  }
  return true;
}

void bn_detRel_and::sum_out(const tuint Nas, const tuint*const IDas, bn_rv_base* rv, tdoublePtr& resVec)
{
  bool has_false = false;
  if (rv==this) {
    has_false = !eval();
    for (tuint i=0;i<Nas;++i) {
      const bool b2 = (!has_false && IDas[i]==1) || (has_false && IDas[i]==0);
      if (!b2) resVec[i] = ZERO;
    }
    return;
  } else {
    const tuint pN = get_parentN();
    for (tuint i=0;i<pN;++i) {
      bn_rv_base* rvt = get_parent(i).rv;
      if (rvt==rv) continue;
      if (stateIDs[rvt->get_ID()]==0) {
	has_false = true;
	break;
      }
    }
    if ( (has_false && stateIDs[get_ID()]==0) || (!has_false && stateIDs[get_ID()]==1) ) return;
    for (tuint i=0;i<Nas;++i) {
      resVec[i] = ZERO;
    }
    return;
  }
}

const bool bn_detRel_and::rnd_state(tdouble& p) const
{
  const tuint cS = (eval()?1:0);
  switch (get_obsrvType()) {
    case noObsrv:
    {
      stateIDs[get_ID()] = cS;
      return true;
    }
    case snglObsrv:
      if (stateIDs[get_ID()]==cS) {
	return true;
      } else {
	p=ZERO;
	return false;
      }
    default:
      #if FLX_DEBUG
	throw FlxException_Crude("bn_detRel_and::rnd_state");
      #else
	return true;
      #endif
  }
}

void bn_detRel_and::rnd_state(tdouble& p, tdouble& p2) const
{
  rnd_state(p);
}


bn_dec::bn_dec(const std::string name, const bool outputRV, const std::vector< std::string >& states, bn_node*const node)
: bn_rv_base(name,outputRV,states,std::vector<bn_rv_base*>(),node)
{
  
}

void bn_dec::observe(const std::vector< tuint > oV)
{
  if (oV.size()!=1) {
    std::ostringstream ssV;
    ssV << "A decission can only be observed by a single observation (decission).";
    throw FlxException("bn_dec::observe", ssV.str() );
  }
  flxbn::bn_rv_base::observe(oV);
}

bn_comp::bn_comp(const bool sumb, bn_rv_base*const rv)
: sum_id(rv==NULL?4:(sumb?((rv->get_type()==detRel)?3:((rv->get_obsrvType()==noObsrv)?1:2)):0)),rv(rv),ut(NULL)
{
  
}

bn_comp::bn_comp(bn_utility*const ut)
: sum_id(0),rv(NULL),ut(ut)
{

}

const bool bn_dec::rnd_state(tdouble& p) const
{
  check_valid();
  return true;
}

void bn_dec::rnd_state(tdouble& p, tdouble& p2) const
{
  check_valid();
}

const tdouble bn_dec::get_prob_const() const
{
  check_valid();
  return ONE;
}

void bn_dec::check_valid() const
{
  if (get_obsrvType() != snglObsrv) {
    throw FlxException("bn_dec::check_valid_1", "Decission node '" + name + "' has not been observed.");
  }
  #if FLX_DEBUG
    if (get_observedStatesN()!=1) throw FlxException_Crude("bn_dec::check_valid_2");
  #endif
}

bn_utility::bn_utility(const std::string name, const std::vector< bn_rv_base* > drvs, const flxVec& utilityVec, bn_node*const node)
: name(name), parentN(drvs.size()), node(node), tableCols(1), tableColsVec(drvs.size()), utilityVec(utilityVec)
{
  // set-up parents
    for (tuint i=0;i<parentN;++i) {
      parents.push_back(rv_edge(drvs[i]));
    }
    for (tuint i=0;i<parentN;++i) {
      tableColsVec[parentN-1-i] = tableCols;
      tableCols*=drvs[parentN-1-i]->stateN;
    }
}

inline const tuint bn_utility::get_curr_col() const
{
  const tuint pN = parentN;
  if (pN==0) {
    return 0;
  } else {
    tuint col = 0;
    for (tuint i=0;i<pN;++i) {
      col += tableColsVec[i]*bn_rv_base::stateIDs[parents[i].rv->get_ID()];
    }
    return col;
  }
}

bn_comp* bn_utility::lock_parents(std::vector< Btree_class* >& treevec, bn_comp_ptr& ct_c)
{
  bn_comp* ct = NULL;
  ct_c = NULL;
  // activate information (observations)
    // reset b-tree
      const tuint S2 = treevec.size();
      for (tuint i=0;i<S2;++i) {
	treevec[i]->deactivate();
      }
    for (tuint i=0;i<parentN;++i) {
      const tuint tid = parents[i].rv->get_btreeID()-1;
      if (parents[i].rv->get_obsrvType()!=snglObsrv) {
	if (ct==NULL) {
	  ct_c = new bn_comp(true,parents[i].rv);
	  ct = ct_c;
	} else {
	  ct_c = ct_c->append(parents[i].rv);
	}
      }
      treevec[tid]->set_active();
    }
  for (tuint i=0;i<parentN;++i) {
    parents[i].rv->set_locked(true);
  }
  if (ct==NULL) {
    ct = new bn_comp(false,NULL);	// just a multiplicator
    ct_c = ct;
  }
  ct_c->append(new bn_comp(this));
  return ct;
}

void bn_utility::unlock_parents()
{
  for (tuint i=0;i<parentN;++i) {
    parents[i].rv->set_locked(false);
  }
}

void bn_utility::activate_tree(std::vector< Btree_class* >& treevec) const
{
  // activate base-trees of parents
    for (tuint i=0;i<parentN;++i) {
      const tuint tid = parents[i].rv->get_btreeID()-1;
      treevec[tid]->set_active();
    }
  // lock parents
    for (tuint i=0;i<parentN;++i) {
      parents[i].rv->set_locked(true);
    }
}

void bn_utility::get_parents(std::vector< bn_rv_base* >& pVec) const
{
  if (pVec.empty()==false) {
    throw FlxException_Crude("bn_utility::get_parents");
  }
  for (tuint i=0;i<parentN;++i) {
    pVec.push_back(parents[i].rv);
  }
}


bn_comp::~bn_comp()
{
  for (tuint i=0;i<childs.size();++i) {
    delete childs[i];
  }
}

bn_comp* bn_comp::append(bn_rv_base*const rvs)
{
  bn_comp* t = new bn_comp(true,rvs);
  append(t);
  return t;
}

bn_comp* bn_comp::find_sum(bn_rv_base*const rvs)
{
  if (sum_id>0) {
    if (rvs==rv) {
      return this;
    } else {
      for (tuint i=0;i<childs.size();++i) {
	bn_comp*t = childs[i]->find_sum(rvs);
	if (t) return t;
      }
      return NULL;
    }
  } else {
    return NULL;
  }
}

void bn_comp::append(bn_rv_base*const rvs, const tuint chk_id)
{
  // preliminary test (are decissions observed?)
    if (chk_id==0) {
      if (rvs->get_type()==decNode) {
	bn_dec* bndn = dynamic_cast<bn_dec*>(rvs);
	#if FLX_DEBUG
	  if (bndn==NULL) throw FlxException_Crude("bn_comp::append");
	#endif
	bndn->check_valid();
      }
    }
  if (chk_id>rvs->get_parentN()) {	// append variable here
    bn_comp *t;
    if (rvs->get_type()==detRel) {
      t = new bn_comp(true,rvs);
    } else {
      t = new bn_comp(false,rvs);
    }
    childs.push_back(NULL);
    const tuint childN = childs.size();
    for (tuint i=1;i<childN;++i) childs[childN-i] = childs[childN-i-1];
    childs[0] = t;
    return;
  }
  if (chk_id==0 && rvs->get_obsrvType()==snglObsrv) {
    this->append(rvs,1);
    return;
  }
  bn_comp* t = NULL;
  if (chk_id==0) {
    t = find_sum(rvs);
  } else {
    t = find_sum(rvs->get_parent(chk_id-1).rv);
  }
  if (t==NULL) t = this;
  t->append(rvs,chk_id+1);
}

void bn_comp::append(bn_comp* ct)
{
  #if FLX_DEBUG
    if (sum_id==0) throw FlxException_Crude("bn_comp::append_1");
    if (ct==NULL) throw FlxException_Crude("bn_comp::append_2");
  #endif
  childs.push_back(ct);
}

const tdouble bn_comp::compute() const
{
  switch (sum_id) {
    case 0:
      #if FLX_DEBUG
	if (childs.size()>0) throw FlxException_Crude("bn_comp::compute_1");
      #endif
      if (rv) {
	return rv->get_prob_const();
      } else {
	return ut->get_utility_const();
      }
    case 1:
    {
      const tuint childN = childs.size();
      #if FLX_DEBUG
	if (childN==0) throw FlxException_Crude("bn_comp::compute_2");
      #endif
      const tuint stateN = rv->stateN;
      tuint& stateID = rv->stateIDs[rv->get_ID()];
      const bn_comp* const* childp = &childs[0];
      tdouble res = ZERO;
      for (stateID=0;stateID<stateN;++stateID) {
	tdouble res2 = childp[0]->compute();
	for (tuint i=1;i<childN&&res2>GlobalVar.TOL();++i) {
	  res2 *= childp[i]->compute();
	}
	res += res2;
      }
      return res;
    }
    case 2:
    {
      const tuint childN = childs.size();
      #if FLX_DEBUG
	if (childN==0) throw FlxException_Crude("bn_comp::compute_3");
      #endif
      tuint& stateID = rv->stateIDs[rv->get_ID()];
      const bn_comp* const* childp = &childs[0];
      tdouble res = ZERO;
      const tuint aN = rv->get_observedStatesN();
      for (tuint i=0;i<aN;++i) {
	stateID = rv->get_observedStates(i);
	tdouble res2 = childp[0]->compute();
	for (tuint i=1;i<childN&&res2>GlobalVar.TOL();++i) {
	  res2 *= childp[i]->compute();
	}
	res += res2;
      }
      return res;
    }
    case 3:
    case 4:
    {
      const tuint childN = childs.size();
      const bn_comp* const* childp = &childs[0];
      tdouble res2 = ONE;
      if (sum_id==3 && (!rv->rnd_state(res2))) return ZERO;
      for (tuint i=0;i<childN&&res2>GlobalVar.TOL();++i) {
	res2 *= childp[i]->compute();
      }
      return res2;
    }
    default:
      #if FLX_DEBUG
	throw FlxException_Crude("bn_comp::compute_4");
      #else
	return ZERO;
      #endif
  }
}

void bn_comp::write(std::ostream& sout) const
{
  if (sum_id>0) {
    if (sum_id==4) {
      sout << "M";
    } else {
      if (sum_id==3) {
	sout << "D";
      } else {
	sout << "S";
      }
      sout << "::" << rv->name;
    }
    sout << "(";
      for (tuint i=0;i<childs.size();++i) {
	if (i>0) sout << "*";
	childs[i]->write(sout);
      }
    sout << ")";
  } else {
    if (rv) {
      sout << rv->name;
    } else {
      sout << "U(" << ut->get_name() << ")";
    }
  }
}

bn_elim_el_table_base::bn_elim_el_table_base(bn_rv* rv)
: Nvars(rv->get_parentN()+1), vars(new bn_rv_base*[Nvars]), Nstates(new tuint[Nvars]), mapTable(new tuint[Nvars]),
curID(new tuint*[Nvars])
{
  // rvs
    vars[0] = rv;
    for (tuint i=1;i<Nvars;++i) vars[i] = rv->get_parent(i-1).rv;
  // states of rvs
    for (tuint i=0;i<Nvars;++i) Nstates[i] = vars[i]->stateN;
  // id's for table
    tuint tableCols = 1;
    for (tuint i=0;i<Nvars;++i) {
      mapTable[Nvars-1-i] = tableCols;
      tableCols *= Nstates[Nvars-1-i];
    }
  // curIDs
    for (tuint i=0;i<Nvars;++i) curID[i] = &(bn_rv_base::stateIDs[vars[i]->get_ID()]);
}

bn_elim_el_table_base::bn_elim_el_table_base()
: Nvars(0), vars(NULL), Nstates(NULL), mapTable(NULL), curID(NULL)
{
  
}

bn_elim_el_table_base::~bn_elim_el_table_base()
{
  if (vars) delete [] vars;
  if (Nstates) delete [] Nstates;
  if (mapTable) delete [] mapTable;
  if (curID) delete [] curID;
}

const tdouble flxbn::bn_elim_el_table_base::get_prob_v(const tdouble*const table) const
{
  tuint IDbase = 0;
  for (tuint i=0;i<Nvars;++i) {
    IDbase += mapTable[i]*(*curID[i]);
  }
  return table[IDbase];
}

const bool bn_elim_el_table_base::depends_on(bn_rv_base* rv)
{
  for (tuint i=0;i<Nvars;++i) {
    if (vars[i] == rv) return true;
  }
  return false;
}

void bn_elim_el_table_base::get_active_RVs(std::vector< bn_rv_base* >& arvlist)
{
  for (tuint i=0;i<Nvars;++i) {
    bn_rv_base* rvt = vars[i];
    const tuint S  = arvlist.size();
    bool notfound = true;
    for (tuint j=0;j<S&&notfound;++j) {
      if (rvt == arvlist[j]) notfound = false;
    }
    if (notfound) arvlist.push_back(rvt);
  }
}

const tdouble bn_elim_el_detRel::get_prob() const
{
  return detRel->get_prob_const();
}

void bn_elim_el_detRel::get_active_RVs(std::vector< bn_rv_base* >& arvlist)
{
  const tuint pN = detRel->get_parentN();
  std::vector< bn_rv_base* > tl;
  tl.push_back(detRel);
  for (tuint i=0;i<pN;++i) tl.push_back(detRel->get_parent(i).rv);
  const tuint Nvars = tl.size();
  for (tuint i=0;i<Nvars;++i) {
    bn_rv_base* rvt = tl[i];
    const tuint S  = arvlist.size();
    bool notfound = true;
    for (tuint j=0;j<S&&notfound;++j) {
      if (rvt == arvlist[j]) notfound = false;
    }
    if (notfound) arvlist.push_back(rvt);
  }
}

bn_elim_el_table_rv::bn_elim_el_table_rv(bn_rv* rv)
: bn_elim_el_table_base(rv), table(rv->get_table_ptr())
{
  
}

void bn_elim_el_table_rv::sum_out(const tuint Nas, const tuint*const IDas, bn_rv_base* rv, tdoublePtr& resVec)
{
  tuint rvidj = 0;
  #if FLX_DEBUG
    bool test_bool = false;
  #endif
  tuint IDbase = 0;
  for (tuint i=0;i<Nvars;++i) {
    if (vars[i]==rv) {	// remember its ID
      rvidj = mapTable[i];
      #if FLX_DEBUG
	test_bool = true;
      #endif
      continue;
    }
    IDbase += mapTable[i]*(*curID[i]);
  }
  #if FLX_DEBUG
    if (!test_bool) throw FlxException_Crude("bn_elim_el_table_rv::sum_out");
  #endif
  for (tuint i=0;i<Nas;++i) {
    resVec[i] *= table[IDbase + rvidj*IDas[i]];
  }
}

const tdouble bn_elim_el_table_rv::get_prob() const
{
  return get_prob_v(table);
}

bn_elim_el_table::bn_elim_el_table(bn_rv_base* rv, std::list< bn_elim_el_base* >& subLst)
: bn_elim_el_table_base(), tableSize(0), table(NULL), deleteTable(true)
{
  // check for plausibility
    #if FLX_DEBUG
      if (subLst.empty()) {
	throw FlxException_Crude("bn_elim_el_table::bn_elim_el_table");
      }
    #endif
  // find all active random variables
    std::vector<bn_rv_base*> arvlist;	// list all active random variables
    arvlist.push_back(rv);
    for (std::list<bn_elim_el_base*>::iterator it = subLst.begin(); it!=subLst.end();++it) {
      bn_elim_el_base* el = (*it);
      el->get_active_RVs(arvlist);
    }
    arvlist.erase(arvlist.begin());	// this RV is summed out
    // remove single observations
    { 
      bool b=true;
      while (b) {
	b = false;
	for (std::vector<bn_rv_base*>::iterator it = arvlist.begin(); it!=arvlist.end();++it) {
	  bn_rv_base* el = (*it);
	  if (el->get_obsrvType()==snglObsrv) {
	    arvlist.erase(it);
	    b = true;
	    break;
	  }
	}
      }
    }
  // allocate memory
    Nvars = arvlist.size();
    tuint tableCols = 1;
    if (Nvars>0) {
      vars = new bn_rv_base*[Nvars];
      Nstates = new tuint[Nvars];
      mapTable = new tuint[Nvars];
      curID = new tuint*[Nvars];
    // rvs
      for (tuint i=0;i<Nvars;++i) {
	vars[i] = arvlist[i];
      }
    // states of rvs
      for (tuint i=0;i<Nvars;++i) Nstates[i] = vars[i]->stateN;
    // id's for table
      for (tuint i=0;i<Nvars;++i) {
	mapTable[Nvars-1-i] = tableCols;
	tableCols *= Nstates[Nvars-1-i];
      }
    // curIDs
      for (tuint i=0;i<Nvars;++i) curID[i] = &(bn_rv_base::stateIDs[vars[i]->get_ID()]);
    }
    table = new tdouble[tableCols];
    tableSize = tableCols;
  // start with the computational part
    // count relevant states in rv
      tuint Nas = 0;
      tuint *IDas = NULL;
      if (rv->get_obsrvType()==noObsrv) {
	const tuint Ns = rv->stateN;
	Nas = Ns;
	IDas = new tuint[Nas];
	for (tuint i=0;i<Nas;++i) {
	  IDas[i] = i;
	}
      } else {
	const tuint No = rv->get_observedStatesN();
	Nas = No;
	IDas = new tuint[Nas];
	for (tuint i=0;i<No;++i) {
	  IDas[i] = rv->get_observedStates(i);
	}
      }
      tdouble* resVec = new tdouble[Nas];
      for (tuint i=0;i<Nas;++i) resVec[i] = ONE;
    // sum out rv
      if (Nvars==0) {
	for (std::list<bn_elim_el_base*>::iterator it = subLst.begin(); it!=subLst.end();++it) {
	  bn_elim_el_base* el = (*it);
	  el->sum_out(Nas,IDas,rv,resVec);
	}
	tdouble res = ZERO;
	for (tuint i=0;i<Nas;++i) res += resVec[i];
	table[0] = res;
      } else {
	tuint tc = 0;	// table counter
	#ifndef FLX_CV_1
	  tuint counter[Nvars];
	#else
	  tuint *counter = new tuint[Nvars];
	#endif
	for (tuint i=0;i<Nvars;++i) counter[i] = 0;
	for (tuint i=0;i<Nvars;++i) *(curID[i]) = 0;	// pre-set all state-IDs
	tuint &ir = counter[Nvars-1];
	while (true) {
	  // increase counter by one
	    tuint ll=0;		// last level (update stateIDs)
	    for (tuint i=0;i<Nvars;++i) {
	      if (counter[Nvars-1-i]>=Nstates[Nvars-1-i]) {
		if (i==Nvars-1) {
		  break;
		} else {
		  counter[Nvars-1-i] = 0;
		  ++counter[Nvars-2-i];
		  continue;
		}
	      } else {
		ll = i;
		break;
	      }
	    }
	    if (counter[0]>=Nstates[0]) break;
	  // update all IDs having changed
	    for (tuint i=0;i<=ll;++i) {	
	      *(curID[Nvars-1-i]) = counter[Nvars-1-i];
	    }
	  // compute probability-coefficient of new table
	    for (std::list<bn_elim_el_base*>::iterator it = subLst.begin(); it!=subLst.end();++it) {
	      bn_elim_el_base* el = (*it);
	      el->sum_out(Nas,IDas,rv,resVec);
	    }
	    tdouble res = ZERO;
	    for (tuint i=0;i<Nas;++i) {
	      res += resVec[i];
	      resVec[i] = ONE;
	    }
	    table[tc] = res;
	  ++tc;		// next coefficient
	  ++ir;		// actually increase counter
	}
	#ifdef FLX_CV_1
	  delete [] counter;
	#endif
      }
    delete [] IDas;
    delete [] resVec;
    while (subLst.empty()==false) {
      delete *(subLst.begin());
      subLst.pop_front();
    }
}

bn_elim_el_table::bn_elim_el_table(std::list< bn_elim_el_base* >& subLst)
: bn_elim_el_table_base(), tableSize(0), table(NULL), deleteTable(true)
{
  // check for plausibility
    #if FLX_DEBUG
      if (subLst.empty()) {
	throw FlxException_Crude("bn_elim_el_table::bn_elim_el_table_1");
      }
    #endif
  // find all active random variables
    std::vector<bn_rv_base*> arvlist;	// list all active random variables
    for (std::list<bn_elim_el_base*>::iterator it = subLst.begin(); it!=subLst.end();++it) {
      bn_elim_el_base* el = (*it);
      el->get_active_RVs(arvlist);
    }
  // allocate memory
    Nvars = arvlist.size();
    tuint tableCols = 1;
    if (Nvars>0) {
      vars = new bn_rv_base*[Nvars];
      Nstates = new tuint[Nvars];
      mapTable = new tuint[Nvars];
      curID = new tuint*[Nvars];
    // rvs
      for (tuint i=0;i<Nvars;++i) {
	vars[i] = arvlist[i];
      }
    // states of rvs
      for (tuint i=0;i<Nvars;++i) Nstates[i] = vars[i]->stateN;
    // id's for table
      for (tuint i=0;i<Nvars;++i) {
	mapTable[Nvars-1-i] = tableCols;
	tableCols *= Nstates[Nvars-1-i];
      }
    // curIDs
      for (tuint i=0;i<Nvars;++i) curID[i] = &(bn_rv_base::stateIDs[vars[i]->get_ID()]);
    } else {
      throw FlxException_Crude("bn_elim_el_table::bn_elim_el_table_2");
    }
    table = new tdouble[tableCols];
    tableSize = tableCols;
    for (tuint i=0;i<tableSize;++i) table[i] = ONE;
  // start with the computational part  (this might be done much faster)
    tuint tc = 0;	// table counter
    #ifndef FLX_CV_1
      tuint counter[Nvars];
    #else
      tuint *counter = new tuint[Nvars];
    #endif
    for (tuint i=0;i<Nvars;++i) counter[i] = 0;
    for (tuint i=0;i<Nvars;++i) *(curID[i]) = 0;	// pre-set all state-IDs
    tuint &ir = counter[Nvars-1];
    while (true) {
      // increase counter by one
	tuint ll=0;		// last level (update stateIDs)
	for (tuint i=0;i<Nvars;++i) {
	  if (counter[Nvars-1-i]>=Nstates[Nvars-1-i]) {
	    if (i==Nvars-1) {
	      break;
	    } else {
	      counter[Nvars-1-i] = 0;
	      ++counter[Nvars-2-i];
	      continue;
	    }
	  } else {
	    ll = i;
	    break;
	  }
	}
	if (counter[0]>=Nstates[0]) break;
      // update all IDs having changed
	for (tuint i=0;i<=ll;++i) {	
	  *(curID[Nvars-1-i]) = counter[Nvars-1-i];
	}
      // compute probability-coefficient of new table
	tdouble res = ONE;
	for (std::list<bn_elim_el_base*>::iterator it = subLst.begin(); it!=subLst.end();++it) {
	  bn_elim_el_base* el = (*it);
	  res *= el->get_prob();
	}
	table[tc] = res;	
      ++tc;		// next coefficient
      ++ir;		// actually increase counter
    }
    while (subLst.empty()==false) {
      delete *(subLst.begin());
      subLst.pop_front();
    }
    #ifdef FLX_CV_1
      delete [] counter;
    #endif
}

bn_elim_el_table::bn_elim_el_table(const tdouble p)
: bn_elim_el_table_base(), tableSize(1), table(new tdouble[1]), deleteTable(true)
{
  *table = p;
}

bn_elim_el_table::~bn_elim_el_table()
{
  if (deleteTable) if (table) delete [] table;
}

void bn_elim_el_table::sum_out(const tuint Nas, const tuint*const IDas, bn_rv_base* rv, tdoublePtr& resVec)
{
  tuint rvidj = 0;
  tuint IDbase = 0;
  for (tuint i=0;i<Nvars;++i) {
    if (vars[i]==rv) {	// remember its ID
      rvidj = mapTable[i];
      continue;
    }
    IDbase += mapTable[i]*(*curID[i]);
  }
  for (tuint i=0;i<Nas;++i) {
    resVec[i] *= table[IDbase + rvidj*IDas[i]];
  }
}

const tdouble bn_elim_el_table::get_prob() const
{
  return get_prob_v(table);
}

const tdouble bn_elim_el_table::exp_Utility_of_table(const bn_utility& utl) const
{
  if (Nvars==0) {	// there might be utility nodes
    return utl.get_utility_const();
  }
  tdouble ExpUt = ZERO;
  tuint tc = 0;		// table counter
  #ifndef FLX_CV_1
    tuint counter[Nvars];
  #else
    tuint *counter = new tuint[Nvars];
  #endif
  for (tuint i=0;i<Nvars;++i) counter[i] = 0;
  for (tuint i=0;i<Nvars;++i) *(curID[i]) = 0;	// pre-set all state-IDs
  tuint &ir = counter[Nvars-1];
  while (true) {
    // increase counter by one
      tuint ll=0;		// last level (update stateIDs)
      for (tuint i=0;i<Nvars;++i) {
	while (		// TODO more efficient && consider det-nodes
	  (*(vars[Nvars-1-i])).get_obsrvType()!=noObsrv 
	  && counter[Nvars-1-i]<Nstates[Nvars-1-i] 
	  && (*(vars[Nvars-1-i])).is_state_observed(counter[Nvars-1-i])
	) {
	  ++counter[Nvars-1-i];
	}
	if (counter[Nvars-1-i]>=Nstates[Nvars-1-i]) {
	  if (i==Nvars-1) {
	    break;
	  } else {
	    counter[Nvars-1-i] = 0;
	    ++counter[Nvars-2-i];
	    continue;
	  }
	} else {
	  ll = i;
	  break;
	}
      }
      if (counter[0]>=Nstates[0]) break;
    // update all IDs having changed
      for (tuint i=0;i<=ll;++i) {	
	*(curID[Nvars-1-i]) = counter[Nvars-1-i];
      }
    // compute probability-coefficient of new table
      ExpUt += table[tc]*utl.get_utility_const();
    ++tc;		// next coefficient
    ++ir;		// actually increase counter
  }
  #ifdef FLX_CV_1
    delete [] counter;
  #endif
  return ExpUt;
}


Btree_class::Btree_class(const int ID, bn_rv_base* const root, const std::vector<bn_rv_base*>& rvid)
: active(false),prob(ZERO),ID(ID),root(root)
{
  // create list with all RVs in base tree
    const tuint S = rvid.size();
    for (tuint i=0;i<S;++i) {
      bn_rv_base* rvt = rvid[i];
      if (rvt->get_btreeID()==ID) btrvs.push_back(rvt);
    }
}

void Btree_class::set_prob()
{
  // calculate probability
    prob = bnm.calc_prob(ID);
}

void flxbn::Btree_class::clear_topo_sub()
{
  // reset sub-trees
    const tuint rvN = btrvs.size();
    for (tuint i=0;i<rvN;++i) {
      btrvs[i]->clear_topo(false);	// clear topo of sub-trees
    }
}

Btree_class& Btree_class::operator=(const flxbn::Btree_class& rhs)
{
  throw FlxException_Crude("Btree_class::operator=");
}

void Btree_class::assemble_bList(std::list< bn_rv_base* >& curL) const
{
  const tuint S = btrvs.size();
  for (tuint i=0;i<S;++i) {
    curL.push_back(btrvs[i]);
  }
}

void bn_manager::update_observedStateValues()
{
  const tuint S = rvid.size();
  if (bn_rv::stateIDs) {
    delete [] bn_rv::stateIDs;
  }
  bn_rv::stateIDs = new tuint[S];
  for (tuint i=0;i<S;++i) {
    if (rvid[i]->get_obsrvType()==snglObsrv) {
      bn_rv::stateIDs[i] = rvid[i]->get_observedStates(0);
    } else {
      bn_rv::stateIDs[i] = 0;
    }
  }
}

void bn_manager::compute_shortestPath(const tuint strvsS, tuint* distRV, tuint* pathRV, bool* strv_active, std::vector< bn_rv_base* > strvs)
{
  // get shortest path
    // initialize distance matrix
      for (tuint i=0;i<strvsS;++i) {
	if (!strv_active[i]) continue;
	bn_rv_base* trv = strvs[i];
	for (tuint j=0;j<strvsS;++j) {
	  if (!strv_active[j]) continue;
	  distRV[i*strvsS+j] = trv->get_dist(strvs[j]);
	  pathRV[i*strvsS+j] = std::numeric_limits<tuint>::max();
	}
      }
    // run Floyd–Warshall_algorithm
      for (tuint k=0;k<strvsS;++k) {
	if (!strv_active[k]) continue;
	for (tuint i=0;i<strvsS;++i) {
	  if (!strv_active[i]) continue;
	  for (tuint j=0;j<strvsS;++j) {
	    if (!strv_active[j]) continue;
	    if (distRV[i*strvsS+k]==std::numeric_limits<tuint>::max()) continue;
	    if (distRV[k*strvsS+j]==std::numeric_limits<tuint>::max()) continue;
	    const tuint pd = distRV[i*strvsS+k]+distRV[k*strvsS+j];
	    if (pd<distRV[i*strvsS+j]) {
	      distRV[i*strvsS+j] = pd;
	      pathRV[i*strvsS+j] = k;
	    }
	  }
	}
      }
}

const bool bn_manager::check_shortestPath(const tuint strvsS, tuint* distRV, tuint* pathRV, bool* strv_active, std::vector< bn_rv_base* > strvs, const tuint start, const tuint end) const
{
  #if FLX_DEBUG
    if (distRV[start*strvsS+end]==std::numeric_limits<tuint>::max()) throw FlxException_Crude("bn_manager::check_shortestPath");
  #endif
  const tuint intermediate = pathRV[start*strvsS+end];
  if (intermediate==std::numeric_limits<tuint>::max()) {
    return true;
  } else {
    if (!strv_active[intermediate]) {
      return false;
    }
  }
  if (!check_shortestPath(strvsS,distRV,pathRV,strv_active,strvs,start,intermediate)) return false;
  return check_shortestPath(strvsS,distRV,pathRV,strv_active,strvs,intermediate,end);
}

const tdouble bn_manager::reduce_btree(std::list< bn_rv_base* >& curL)
{
  tdouble p=ONE;
  // output
    GlobalVar.slog(4) << "  compute probability on tree: ";
    output_list(GlobalVar.slog(4),curL);
    GlobalVar.slog(4) << std::endl;
  std::list<bn_rv_base*>::reverse_iterator rit;
  for (bool bwatch=true;bwatch;) {				// as long as there is change !!!
    bwatch = false;
    for (rit=curL.rbegin(); rit!=curL.rend(); ++rit) {
      bn_rv_base* rvt = (*rit);
      switch (rvt->get_obsrvType()) {
	case noObsrv:
	  if (!rvt->is_locked()) {
	    const tuint CNC = rvt->get_childN();
	    tuint cn = CNC;
	    for (tuint i=0;i<CNC;++i) {
	      if (rvt->get_child(i).get_streeID()==-2) --cn;
	    }
	    if (cn==0) {
	      rvt->stree_remove();
	      curL.erase((++rit).base());
	      bwatch=true;
	    }
	  }
	  break;
	case snglObsrv:		// it's a single observation
	{
	  if (!rvt->is_locked()) {
	    const tuint PNC = rvt->get_parentN();
	    tuint pn = PNC;
	    for (tuint i=0;i<PNC;++i) {
	      bn_rv_base* rvt2 = rvt->get_parent(i).rv;
	      if (rvt2->get_obsrvType()==snglObsrv && !(rvt2->is_locked())) --pn;
	    }
	    if (pn==0) {
	      p*=rvt->get_prob_const();
	      curL.erase((++rit).base());
	      bwatch=true;
	    }
	  }
	  break;
	}
	case multObsrv:		// it's a multiple observation
	{
	  break;
	}
	default:
	  #if FLX_DEBUG
	    throw FlxException_Crude("bn_manager::reduce_btree");
	  #endif
	  break;
      }
      if (bwatch) break;
    }
  }
  // output
    GlobalVar.slog(4) << "  ... after reduction: ";
    output_list(GlobalVar.slog(4),curL);
    GlobalVar.slog(4) << std::endl;
  return p;
}

bn_comp* bn_manager::extract_tree(const int btreeID, std::list< bn_rv_base* >& curL)
{
  bn_comp* ct_root = new bn_comp(false,NULL);
  try {
  // find sub-trees
    std::list<bn_rv_base*>::iterator it;
    std::list<bn_rv_base*>::iterator itb = curL.begin();	// current beginning
    const tuint rvidS = rvid.size();
    int streeID = 0;
    while (true) {
      bool brid = false;
      for (it=itb;it!=curL.end();++it) {	// find potential root of sub-tree
	if ((*it)->get_streeID()==0) {
	  if ((*it)->get_obsrvType()==snglObsrv) {
	    if ((*it)->is_locked()) {
	      ct_root->append(new bn_comp(false,(*it)));
	      continue;
	    } else {
	      throw FlxException_Crude("bn_manager::calc_prob_sum_2");
	    }
	  }
	  brid = true;
	  itb = it;
	  break;
	}
      }
      if (!brid) break;	// all S-trees assembled
      // detect topology of sub-tree
      {
	++streeID;
	rvid[(*it)->get_ID()]->detect_topo_sub(btreeID,streeID,0,NULL);
	has_topology=true;
	#ifndef FLX_CV_1
	  bool vist[rvidS];		// variables in sub-tree
	  int vist_id[rvidS];			// mapping of global_id to subtree_id
	#else
	  bool *vist = new bool[rvidS];		// variables in sub-tree
	  int *vist_id = new int[rvidS];			// mapping of global_id to subtree_id
	#endif
	  for (tuint i=0;i<rvidS;++i) vist[i] = false;
	rvid[(*it)->get_ID()]->get_subtree_rvs(vist,streeID);
	correct_vistVec(vist,curL);
	std::vector<bn_rv_base*> strvs;	// sub-tree RVs
	for (tuint i=0;i<rvidS;++i) {
	  if (vist[i]) {
	    vist_id[i] = strvs.size();
	    strvs.push_back(rvid[i]);
	  } else {
	    vist_id[i] = -2;
	  }
	}
	const tuint strvsS = strvs.size();
	#ifndef FLX_CV_1
	  bool strv_active[strvsS];		// currently active RVs
	  int strv_indic[strvsS];		// idicator-value for the RVs
	  tuint distRV[strvsS*strvsS];
	  tuint pathRV[strvsS*strvsS];
	#else
	  bool *strv_active = new bool[strvsS];		// currently active RVs
	  int *strv_indic = new int[strvsS];		// idicator-value for the RVs
	  tuint *distRV = new tuint[strvsS*strvsS];
	  tuint *pathRV = new tuint[strvsS*strvsS];
	#endif
	  for (tuint i=0;i<strvsS;++i) strv_active[i] = true;
	  for (tuint i=0;i<strvsS;++i) strv_indic[i] = 0;
	// output
	  GlobalVar.slog(4) << "  ... assemble computation sub-tree for: ";
	  std::list<bn_rv*>::reverse_iterator rit;
	  for (tuint i=0;i<strvsS;++i) {
	    if (i>0) GlobalVar.slog(4) << ", ";
	    GlobalVar.slog(4) << strvs[i]->name; 
	    if (strvs[i]->get_obsrvType()==snglObsrv) GlobalVar.slog(4) << "!";
	    else if (strvs[i]->get_obsrvType()==multObsrv ) GlobalVar.slog(4) << "°";
	  }
	  GlobalVar.slog(4) << std::endl;
	// compute distance-matrix
	  compute_shortestPath(strvsS,distRV,pathRV,strv_active,strvs);
	bn_comp* ct = NULL;	// the computation tree
	std::list<bn_comp*> ctl;
	int c = 0;
	tuint rvidC = 0;
	try {
	while (true) {
	  // deactivate observations AND locked variables
	    for (tuint i=0;i<strvsS;++i) {
	      if (strv_active[i]) {
		if (strvs[i]->get_obsrvType()==snglObsrv||strvs[i]->get_type()!=trueRV||strvs[i]->is_locked()) {
		  strv_indic[i] = -1;
		  strv_active[i] = false;
		}
	      }
	    }
	  // compute averages
// 	    GlobalVar.slog(4) << "    distances: ";
	    tuint min_avg_value = std::numeric_limits<tuint>::max();
	    tuint min_avg_id = 0;
	    for (tuint i=0;i<strvsS;++i) {
	      if (!strv_active[i]) continue;
	      tuint tmp_avg = 0;
	      tuint *ip = &distRV[i*strvsS];
	      for (tuint j=0;j<strvsS;++j,++ip) {
		if (!strv_active[j]) continue;
		tmp_avg += *ip;
	      }
// 	      if (min_avg_value<std::numeric_limits<tuint>::max()) GlobalVar.slog(4) << ", ";
// 	      GlobalVar.slog(4) << strvs[i]->name << "(" << tmp_avg << ")";
	      if (tmp_avg<min_avg_value) {
		min_avg_value = tmp_avg;
		min_avg_id = i;
	      }
	    }
	    // only observations ... ???
	      if (min_avg_value == std::numeric_limits<tuint>::max()) break;
// 	    GlobalVar.slog(4) << std::endl;
	    #if FLX_DEBUG
	      if (strv_active[min_avg_id]==false) throw FlxException_Crude("bn_manager::calc_prob_sum_3");
	    #endif
	    if (ct==NULL) {
	      ct = new bn_comp(true,strvs[min_avg_id]);
	      ctl.push_back(ct);
	    } else {
	      std::list<bn_comp*>::iterator ctl_it = ctl.end();
	      ctl_it--;
	      ctl.push_back((*ctl_it)->append(strvs[min_avg_id]));
	    }
	    strvs[min_avg_id]->stree_remove();
	    strv_indic[min_avg_id] = -2;
	    strv_active[min_avg_id] = false;
	    bool found = false;
	    do {
	      for (tuint i=0;i<strvsS;++i) {
		if (strv_indic[i] == c) {
		  found = true;
		  rvidC = i;
		  ++c;
		  break;
		}
		#if FLX_DEBUG
		else if (strv_indic[i] > c) throw FlxException_Crude("bn_manager::calc_prob_sum_4");
		#endif
	      }
	      if (!found) {
		if (c==0) break;
		else {
		  ctl.pop_back();
		  #if FLX_DEBUG
		    if (ctl.size()!=tuint(c)) throw FlxException_Crude("bn_manager::calc_prob_sum_5");
		  #endif
		  --c;
		}
	      }
	    } while (found==false);
	    if (found==false) {
	      break;
	    }
	    // get active variables in new sub-tree
	      for (tuint i=0;i<rvidS;++i) vist[i] = false;
	      for (tuint i=0;i<strvsS;++i) strv_active[i] = false;
	      strvs[rvidC]->get_subtree_rvs(vist,streeID);
	      correct_vistVec(vist,curL);
	      for (tuint i=0;i<rvidS;++i) {
		if (vist[i] && rvid[i]->get_streeID()!=-2) {
		  strv_active[vist_id[i]] = true;
		  if (!(rvid[i]->get_obsrvType()==snglObsrv)) {
		    #if FLX_DEBUG
		      if ((rvid[i]->get_type()==trueRV&&(!rvid[i]->is_locked())) && (strv_indic[vist_id[i]]!=c-1)) throw FlxException_Crude("bn_manager::calc_prob_sum_6");
		    #endif
		    strv_indic[vist_id[i]] = c;
		  }
		}
	      }
	    // check RV-distance
	      bool do_update = false;
	      for (tuint i=0;i<strvsS&&!do_update;++i) {
		if (!strv_active[i]) continue;
		for (tuint j=0;j<strvsS&&!do_update;++j) {
		  if (!strv_active[j]) continue;
		  if (!check_shortestPath(strvsS,distRV,pathRV,strv_active,strvs,i,j)) {
		    do_update = true;
		  }
		}
	      }
	      if (do_update) {
		GlobalVar.slog(4) << "    ... update distance matrix" << std::endl;
		compute_shortestPath(strvsS,distRV,pathRV,strv_active,strvs);
	      }
	}
	// 1. get variables in sub-tree
	//      sort w.r.t. ID
	//      get indication-vector -> initialize 0  (c=0)
	// 1. get distance-matrix
	// 1. deactivate observations  -> -1
	// LOOP N
	// N. find least average node
	//   remove from sub-tree
	//   add as sum-variable in computation-tree
	//   deactivate variable  -> -2
	// N. find first not deactivated RV  which has c! 
	//    if non found 
	//	if c==0  -> BREAK N
	//	else c--
	//    else
	//	c++
	// N. get variables in sub-tree -> assign all of them c
	// N. for every active child, check all path-variables are active
	//    if not: recompute distance matrix for active part
	// N. end loop
	if (ct==NULL) {		// -> no summation variables in sub-tree
	  ct = new bn_comp(false,NULL);
	}
	for (tuint i=0;i<strvsS;++i) {
	  ct->append(strvs[i],0);
	}
	ct_root->append(ct);
	ct = NULL;
	} catch (FlxException &e) {
	  FLXMSG("bn_manager::calc_prob_sum_7",1);
	  if (ct) delete ct;
	  throw;
	}
	#ifdef FLX_CV_1
	  delete [] distRV;
	  delete [] pathRV;
	  delete [] strv_indic;
	  delete [] strv_active;
	  delete [] vist_id;
	  delete [] vist;
	#endif
      }
    }
  } catch (FlxException &e) {
    FLXMSG("bn_manager::calc_prob_sum_8",1);
    if (ct_root) delete ct_root;
    throw;
  }
  return ct_root;
}

void bn_manager::correct_vistVec(bool* vist, std::list< bn_rv_base* >& curL) const
{
  const tuint S = rvid.size();
  #ifndef FLX_CV_1
    bool b[rvid.size()];
  #else
    bool *b = new bool[rvid.size()];
  #endif
  for (tuint i=0;i<S;++i) b[i] = false;
  for (std::list<bn_rv_base*>::iterator it=curL.begin();it!=curL.end();++it) {	// find potential root of sub-tree
    b[(*it)->get_ID()] = true;
  }
  for (tuint i=0;i<S;++i) {
    if (vist[i] && !b[i]) vist[i] = false;
  }
  #ifdef FLX_CV_1
    delete [] b;
  #endif
}

void bn_manager::extract_tableList(std::list< bn_elim_el_base* >& tblLst, const int btreeID, std::list< bn_rv_base* >& curL)
{
  std::list<bn_rv_base*>::iterator it;
  for (it=curL.begin();it!=curL.end();++it) {
    bn_rv_base* rvt = (*it);
    bn_rv* rvt_rv = dynamic_cast<bn_rv*>(rvt);
    if (rvt_rv) {
      tblLst.push_back(new bn_elim_el_table_rv(rvt_rv));
    } else {
      bn_detRel* rvt_det = dynamic_cast<bn_detRel*>(rvt);
      if (rvt_det) {
	tblLst.push_back(new bn_elim_el_detRel(rvt_det));
      } else {
	bn_dec* rvt_dec = dynamic_cast<bn_dec*>(rvt);
	if (rvt_dec) {
	  rvt_dec->check_valid();
	} else {
	  throw FlxException_Crude("bn_manager::extract_tableList_2");
	}
      }
    }
  }
}

const tdouble bn_manager::calc_prob(const int btreeID)
{
  #if FLX_DEBUG
    if (!has_topology) throw FlxException_Crude("bn_manager::calc_prob_1");
    if (btreeID<=0) throw FlxException_Crude("bn_manager::calc_prob_2");
  #endif
  treevec[btreeID-1]->clear_topo_sub();
  tdouble p=ONE;
  // reduce base tree
    std::list<bn_rv_base*> curL;
    treevec[btreeID-1]->assemble_bList(curL);
    p *= reduce_btree(curL);
  // compute probability (of reduced base-tree)
    if (curL.size()>0) {
      switch (smp_mthd) {
	case 0:
	  p *= calc_prob_sum(btreeID,curL);
	  break;
	case 1:
	  p *= calc_prob_LWsim(btreeID,curL);
	  break;
	case 2:
	  p *= calc_prob_elim(btreeID,curL);
	  break;
	default:
	  throw FlxException_Crude("bn_manager::calc_prob_3");
      }
    }
  GlobalVar.slog(4) << "  ... result = " << GlobalVar.Double2String(p,true) << std::endl;
  return p;
}

const tdouble bn_manager::calc_prob_sum(const int btreeID, std::list< bn_rv_base* >& curL)
{
  #if FLX_DEBUG
    if (btreeID<=0) throw FlxException_Crude("bn_manager::calc_prob_sum_1");
  #endif
  tdouble p = ONE;
  bn_comp* ct = extract_tree(btreeID, curL);
  try {
    // output
      GlobalVar.slog(4) << "    ... combined computation tree: ";
      ct->write(GlobalVar.slog(4));
      GlobalVar.slog(4) << std::endl;
    const tdouble p_tmp = ct->compute();
    GlobalVar.slog(4) << "    ... result of computation tree = " << GlobalVar.Double2String(p_tmp,true) << std::endl;
    p *= p_tmp;
  } catch (FlxException &e) {
    FLXMSG("bn_manager::calc_prob_sum",1);
    delete ct;
    throw;
  }
  delete ct;
  return p;
}

const tdouble bn_manager::calc_prob_elim(const int btreeID, std::list< bn_rv_base* >& curL)
{
  #if FLX_DEBUG
    if (btreeID<=0) throw FlxException_Crude("bn_manager::calc_prob_sum_1");
  #endif
  // create list 
    std::list<bn_elim_el_base*> tblLst;
    extract_tableList(tblLst,btreeID,curL);
  // compute list
    bn_elim_el_table* pt = calc_prob_elim(tblLst);
    #if FLX_DEBUG
      if (tblLst.size()>0) {
	delete pt;
	throw FlxException_Crude("bn_manager::calc_prob_elim_1");
      }
    #endif
    if (pt->has_rvs()) {
      delete pt;
      throw FlxException_Crude("bn_manager::calc_prob_elim_2");
    }
    const tdouble res = *(pt->get_table_link());
    delete pt;
    return res;
}

bn_elim_el_table* bn_manager::calc_prob_elim(std::list< bn_elim_el_base* >& tblLst)
{
  // check completeness of elimination order
    tuint c = 0;
    for (std::vector<bn_rv_base*>::iterator it=rvid.begin();it!=rvid.end();++it) {
      bn_rv_base* rvt = (*it);
      if (rvt->get_type() != decNode) ++c;
    }
    if (elimOrder.size() != c) {
      std::ostringstream ssV;
      ssV << "Elimination order list is not complete, since " << elimOrder.size() << "!=" << c << ".";
      throw FlxException_NeglectInInteractive("bn_manager::calc_prob_elim'_1", ssV.str() );
    }
  // for each variable in the elimination order
    tdouble p = ONE;	// probability value with which to multiply the returned array
    for (std::list<bn_rv_base*>::iterator it=elimOrder.begin();it!=elimOrder.end();++it) {
      bn_rv_base* rvt = (*it);
      if (rvt->is_locked()) continue;	// locked variables are NOT eliminated
      std::list< bn_elim_el_base* > subLst;
      // find relevant elements
	for (std::list<bn_elim_el_base*>::iterator it2=tblLst.begin();it2!=tblLst.end();++it2) {
	  bn_elim_el_base* el = (*it2);
	  if (el->depends_on(rvt)) {
	    subLst.push_back(el);
	  }
	}
      if (subLst.size()>0) {
	// remove relevant elements
	  for (std::list<bn_elim_el_base*>::iterator it2=subLst.begin();it2!=subLst.end();++it2) {
	    bn_elim_el_base* el = (*it2);
	    tblLst.remove(el);
	  }
	// compute the resulting new table .. delete tables in subLst
	  bn_elim_el_table* rt = new bn_elim_el_table(rvt, subLst);
	// check if table contains any random variables -> otherwise update p
	  if (rt->has_rvs()) {
	    tblLst.push_back(rt);
	  } else {
	    p *= (rt->get_table_link())[0];
	    delete rt;
	  }
      }
    }
  // decide what to return
    if (tblLst.empty()) {
      return new bn_elim_el_table(p);
    } else {
      if (tblLst.size()>1) {
	bn_elim_el_table* res = new bn_elim_el_table(tblLst);
	#if FLX_DEBUG
	  if (tblLst.empty()==false) throw FlxException_Crude("bn_manager::calc_prob_elim'_2");
	#endif
	tblLst.push_back(res);
      }
      if (tblLst.size()==1) {
	bn_elim_el_base* tb0 = (*tblLst.begin());
	tblLst.pop_front();
	bn_elim_el_table* res = dynamic_cast<bn_elim_el_table*>(tb0);
	if (res==NULL) {
	  bn_elim_el_table_base* res1 = dynamic_cast<bn_elim_el_table_base*>(tb0);
	  if (res1==NULL || res1->has_rvs()!=1) {
	    throw FlxException_NotImplemented("bn_manager::calc_prob_elim'_3");
	  } else {
	    tblLst.push_back(tb0);
	    res = new bn_elim_el_table(tblLst);
	  }
	}
	// include p
	  tdouble *t = res->get_table_link();
	  const tuint TS = res->get_tableSize();
	  for (tuint i=0;i<TS;++i) {
	    t[i] *= p;
	  }
	return res;
      } else {
	throw FlxException_NotImplemented("bn_manager::calc_prob_elim'_4");
      }
    }
}

const tdouble bn_manager::calc_prob_LWsim(const int btreeID, std::list< bn_rv_base* >& curL)
{
  #if FLX_DEBUG
    if (btreeID<=0) throw FlxException_Crude("bn_manager::calc_prob_LWsim_1");
  #endif
  const tuint N = smp_mthd_2_N;
  tdouble res = ZERO;
  std::list<bn_rv_base*>::iterator it;
  if (smp_mthd_2_eP) {
    tdouble resD = ZERO;
    for (tuint i=0;i<N;++i) {
      tdouble p = ONE;		// counter hits (weighted)
      tdouble p2 = ONE;		// counter trials (weighted)
      for (it=curL.begin();it!=curL.end();++it) {
	(*it)->rnd_state(p,p2);
      }
      res += p;
      resD += p2;
    }
    GlobalVar.slog(4) << "  MCI-hits: " << GlobalVar.Double2String(res,true) << " out of " << GlobalVar.Double2String(resD,true) << std::endl;
    return res/resD;
  } else {
    for (tuint i=0;i<N;++i) {
      tdouble p = ONE;		// counter hits
      for (it=curL.begin();it!=curL.end();++it) {
	if (!((*it)->rnd_state(p))) break;
      }
      res += p;
    }
    GlobalVar.slog(4) << "  MCI-hits:" << GlobalVar.Double2String(res,true) << " out of " << GlobalVar.Double2String(N,true) << std::endl;
    return res/N;
  }
}

const tdouble bn_manager::calc_expUtility_sum()
{
  tdouble ut_sum = ZERO;
  const tuint TS = treevec.size();
  for (tuint i=0;i<utvec.size();++i) {
    bn_utility* cur_ut = utvec[i];
    GlobalVar.slog(4) << " E[" << cur_ut->get_name() << "]" << std::endl;
    bn_comp* ct_c = NULL;	// ct_c=leaf of comp. tree; ctb=root of comp. tree
    bn_comp* ctb = cur_ut->lock_parents(treevec,ct_c);
    tdouble p=ONE;
    tdouble lp = ONE;
    try {
      for (tuint j=0;j<TS;++j) {
	if (treevec[j]->is_active()) {
	  treevec[j]->clear_topo_sub();
	  // reduce base tree
	    std::list<bn_rv_base*> curL;
	    treevec[j]->assemble_bList(curL);
	    p *= reduce_btree(curL);
	  // obtain computation-tree of reduced base-tree
	    if (curL.size()>0) {
	      ct_c->append(extract_tree(j+1, curL));
	    }
	  lp = treevec[j]->get_prob();
	}
      }
      p *= ctb->compute();
      p /= lp;
      GlobalVar.slog(4) << "    ... result of E[" << cur_ut->get_name() << "] = " << GlobalVar.Double2String(p) << std::endl;
    } catch (FlxException &e) {
      FLXMSG("bn_manager::calc_expUtility_sum",1);
      delete ctb;
      throw;
    }
    delete ctb;
    ut_sum += p;
    cur_ut->unlock_parents();
  }
  return ut_sum;
}

const tdouble bn_manager::calc_expUtility_elim()
{
  tdouble ut_sum = ZERO;
  const tuint TS = utvec.size();
  for (tuint i=0;i<TS;++i) {
    bn_utility* cur_ut = utvec[i];
    GlobalVar.slog(4) << " E[" << cur_ut->get_name() << "]" << std::endl;
    std::vector< bn_rv_base* > ilt;
    cur_ut->get_parents(ilt);		// write parents of cur_ut to ilt
    bn_elim_el_table* upt = calc_condProb_Elimination(ilt);
    try {
      const tdouble ut_sum_local = upt->exp_Utility_of_table(*cur_ut);
      GlobalVar.slog(4) << "    ... result of E[" << cur_ut->get_name() << "] = " << GlobalVar.Double2String(ut_sum_local) << std::endl;
      ut_sum += ut_sum_local;
      delete upt;
    } catch (FlxException &e) {
      FLXMSG("bn_manager::calc_expUtility_elim",1);
      delete upt;
      throw;
    }
  }
  return ut_sum;
}

const tdouble bn_manager::calc_expUtility_LWsim()
{
  GlobalVar.slog(4) << "  compute expected utility using simulation methods" << std::endl;
  // activate relevant RVs
    // reset b-tree
      const tuint TS = treevec.size();
      for (tuint i=0;i<TS;++i) {
	treevec[i]->deactivate();
      }
    // activate base-trees of utilities
      const tuint UTS = utvec.size();
      for (tuint i=0;i<UTS;++i) {
	utvec[i]->activate_tree(treevec);
      }
    // reduce active base trees
      std::list<bn_rv_base*> curL;
      for (tuint i=0;i<TS;++i) {
	treevec[i]->clear_topo_sub();
	if (treevec[i]->is_active()) {
	  treevec[i]->assemble_bList(curL);
	}
      }
      reduce_btree(curL);
    // unlock RVs
      for (tuint i=0;i<UTS;++i) {
	utvec[i]->unlock_parents();
      }
  const tuint N = smp_mthd_2_N;
  tdouble res_E = ZERO;
  std::list<bn_rv_base*>::iterator it;
  if (smp_mthd_2_eP) {
    tdouble Nc = ZERO; 	// count of probable hits
    for (tuint i=0;i<N;++i) {
      tdouble p = ONE;		// counter hits
      tdouble p2 = ONE;
      bool probable = true;
      for (it=curL.begin();it!=curL.end()&&probable;++it) {
	(*it)->rnd_state(p,p2);
	if (p<=GlobalVar.TOL()) {
	  probable=false;
	  break;
	}
      }
      if (probable) {
	Nc += p2;
	tdouble ut = ZERO;
	for (tuint i=0;i<UTS;++i) {
	  ut += utvec[i]->get_utility_const();
	}
	res_E += ut*p2;
      }
    }
    GlobalVar.slog(4) << "  ... number of possible hits: " << GlobalVar.Double2String(Nc) << std::endl;
    res_E = res_E/Nc;
    GlobalVar.slog(4) << "  ... result = " << GlobalVar.Double2String(res_E) << std::endl;
    return res_E;
  } else {
    tdouble res_sd = ZERO;
    tuint Nc = 0; 	// count of probable hits
    for (tuint i=0;i<N;++i) {
      tdouble p = ONE;		// counter hits
      bool probable = true;
      for (it=curL.begin();it!=curL.end()&&probable;++it) {
	if (!((*it)->rnd_state(p))) {
	  probable=false;
	  break;
	}
      }
      if (probable) {
	++Nc;
	tdouble ut = ZERO;
	for (tuint i=0;i<UTS;++i) {
	  ut += utvec[i]->get_utility_const();
	}
	res_E += ut;
	res_sd += pow2(ut);
      }
    }
    GlobalVar.slog(4) << "    ... number of possible hits: " << GlobalVar.Double2String(Nc) << std::endl;
    res_sd = sqrt((res_sd-pow2(res_E)/Nc)/(Nc-1));
    GlobalVar.slog(4) << "  ... standard deviation of the utility = " << GlobalVar.Double2String(res_sd) << std::endl;
    res_E = res_E/Nc;
    GlobalVar.slog(4) << "  ... result = " << GlobalVar.Double2String(res_E) << std::endl;
    return res_E;
  }
}

bn_manager::~bn_manager()
{
  if (bn_rv::stateIDs) {
    delete [] bn_rv::stateIDs;
    bn_rv::stateIDs = NULL;
  }
  treevec_delete();
}

void bn_manager::treevec_delete()
{
  for (tuint i=0;i<treevec.size();++i) {
    delete treevec[i];
  }
}

void bn_manager::set_smp_mthd(const tuint smp_mthdV)
{
  GlobalVar.slog(3) << "BN::sampling_method set to ";
  switch (smp_mthdV) {
    case 0:
      GlobalVar.slog(3) << "'recursive conditioning' - exact propagation";
      break;
    case 1:
      GlobalVar.slog(3) << "'likelihood weighting' - stochastic simulation (N=" << GlobalVar.Double2String(smp_mthd_2_N) << ",equalProb=" << (smp_mthd_2_eP?"yes":"no") << ")";
      break;
    case 2:
      GlobalVar.slog(3) << "'variable elimination' - exact propagation";
      break;
    default:
      GlobalVar.slog(3) << std::endl;
      std::ostringstream ssV;
      ssV << "Unknown sampling method (" << smp_mthdV << ")";
      throw FlxException_NeglectInInteractive("bn_manager::set_smp_mthd", ssV.str() );
  }
  GlobalVar.slog(3) << std::endl;
  smp_mthd = smp_mthdV;
}

void bn_manager::add_rv(bn_rv_base* rv)
{
  topo_reset();
  std::pair<std::string, bn_rv_base*>Element(rv->name, rv);
  if ( ! rvs.insert(Element).second ) {;
    std::ostringstream ssV;
    ssV << "BN_RV '" << rv->name << "' has already been defined.";
    throw FlxException("bn_manager::add_rv","BN RV already defined",ssV.str());
  }
  rv->set_ID(rvid.size());
  rvid.push_back(rv);
}

void bn_manager::add_utility(bn_utility* ut)
{
  utvec.push_back(ut);
}

bn_rv_base& bn_manager::get_rv(const std::string& rvn) const
{
  std::map<std::string, bn_rv_base*>::const_iterator pos = rvs.find(rvn);
  if ( pos != rvs.end() ) {
    return (*pos->second);
  } else {
    std::ostringstream ssV_2;
    ssV_2 << "The random variable '" << rvn << "' does not exist.";
    throw FlxException("bn_manager::get_rv", ssV_2.str()); 
  }
}

void bn_manager::create_topo()
{
  if (has_topology) return;
  update_observedStateValues();
  // reset topology
    const tuint rvN = rvid.size();
    for (tuint i=0;i<rvN;++i) {
      rvid[i]->clear_topo();
    }
    treevec_delete();
    treevec.clear();		// delete all base-trees
  tuint rid = 0;
  while (true) {
    // search for root of tree
      bool brid = false;
      for (tuint i=rid;i<rvN;++i) {
	if (rvid[i]->get_btreeID()==0) {
	  brid = true;
	  rid = i;
	  break;
	}
      }
      if (!brid) break;	// all B-trees assembled
    // detect topology of tree
      const int btreeID = treevec.size()+1;
      rvid[rid]->detect_topo(btreeID,0,NULL);
      has_topology=true;
      treevec.push_back(new Btree_class(btreeID,rvid[rid],rvid));
	treevec[treevec.size()-1]->set_prob();
      has_topology=false;
  }
  has_topology=true;
  GlobalVar.slog(4) << "BN::topo (created)" << std::endl;
}

void bn_manager::observe_RV(bn_rv_base* rv, const std::vector< tuint > oV)
{
  if (has_topology && rv->get_type()!=decNode) {
    switch (rv->get_obsrvType()) {
      case noObsrv:
	rv->observe(oV);
	return;
      case snglObsrv:
	if (oV.size()==1) {
	  if (oV[0]==rv->get_observedStates(0)) {
	    return;
	  } else {
	    rv->observe(oV);
	    has_topology = true;
	    treevec[rv->get_btreeID()-1]->set_prob();
	    return;
	  }
	} else {
	  rv->observe(oV);
	  return;
	}
      case multObsrv:
	if (oV.size()==1) {
	  rv->observe(oV);
	  return;
	} else {
	  rv->observe(oV);
	  has_topology = true;
	  treevec[rv->get_btreeID()-1]->set_prob();
	  return;
	}
      default:
	#if FLX_DEBUG
	  throw FlxException_Crude("bn_manager::observe_RV_1");
	#endif
	return;
    }
  } else {
    rv->observe(oV);
    return;
  }
  #if FLX_DEBUG
    throw FlxException_Crude("bn_manager::observe_RV_2");
  #endif
}

void bn_manager::append_elimOrderVar(bn_rv_base* rv)
{
  // check if variable is not alreaday in the list
    std::list<bn_rv_base*>::iterator pos = std::find(elimOrder.begin(),elimOrder.end(),rv);
    if (pos != elimOrder.end()) {	// found
      std::ostringstream ssV_2;
      ssV_2 << "The variable '" << rv->name << "' is alreday part of the elimination list.";
      throw FlxException("bn_manager::append_elimOrderVar_1", ssV_2.str()); 
    }
  // make sure it is not a decission node
    if (rv->get_type() == decNode) {
      std::ostringstream ssV;
      ssV << "Decission nodes cannot be eliminated.";
      throw FlxException_NeglectInInteractive("bn_manager::append_elimOrderVar_2", ssV.str() );
    }
  // append the variable
    elimOrder.push_back(rv);
}

void flxbn::bn_manager::output_topo(std::ostream& sout) const
{
  const tuint S = rvid.size();
  for (tuint i=0;i<S;++i) {
    sout << "  " << rvid[i]->name;
    if (rvid[i]->get_obsrvType()==snglObsrv) {
      sout << "!";
    } else if (rvid[i]->get_obsrvType()==multObsrv) {
      sout << "°";
    }
    sout << "(" ;
    if (rvid[i]->get_parentN()>0) {
      const tuint S2 = rvid[i]->get_parentN();
      sout << "|" ;
      for (tuint j=0;j<S2;++j) {
	if (j>0) sout << ",";
	sout << rvid[i]->get_parent(j).rv->name;
	sout << ":" << rvid[i]->get_parent(j).get_btreeID();
// 	sout << ":" << rvid[i]->get_parent(j).get_streeID();
      }
    }
    if (rvid[i]->get_childN()>0) {
      const tuint S2 = rvid[i]->get_childN();
      sout << "§" ;
      for (tuint j=0;j<S2;++j) {
	if (j>0) sout << ",";
	sout << rvid[i]->get_child(j).rv->name;
	sout << ":" << rvid[i]->get_child(j).get_btreeID();
// 	sout << ":" << rvid[i]->get_child(j).get_streeID();
      }
    }
    sout << ")";
    sout << "\t";
    if (rvid[i]->get_obsrvType()==snglObsrv) {
      sout << "observed(" << rvid[i]->states[rvid[i]->get_observedStates(0)] << ")";
    } else {
      sout << "tree=" << rvid[i]->get_btreeID(); // << ":" << rvid[i]->get_streeID();
    }
    sout << std::endl;
  }
}

void bn_manager::output_observed(std::ostream& sout) const
{
  const tuint S = rvid.size();
  bool f=false;
  for (tuint i=0;i<S;++i) {
    if (rvid[i]->get_obsrvType()!=noObsrv) {
      if (!f) f=true;
      else sout << ',';
      if (rvid[i]->get_obsrvType()==snglObsrv) {
	sout << rvid[i]->name << "=" << rvid[i]->states[rvid[i]->get_observedStates(0)];
      } else {
	sout << rvid[i]->name << '{';
	const tuint S = rvid[i]->get_observedStatesN();
	for (tuint i=0;i<S;++i) {
	  sout << "=" << rvid[i]->get_observedStates(i);
	  if (i<S-1) sout << ";";
	}
	sout << '}';
      }
    }
  }
}

void bn_manager::output_list(std::ostream& sout, const std::list< bn_rv_base* >& curL) const
{
  std::list<bn_rv_base*>::const_iterator it;
  for (it=curL.begin();it!=curL.end();++it) {
    if (it!=curL.begin()) GlobalVar.slog(4) << ", ";
    GlobalVar.slog(4) << (*it)->name; 
    if ((*it)->get_obsrvType()==snglObsrv) GlobalVar.slog(4) << "!";
    else if ((*it)->get_obsrvType()==multObsrv ) GlobalVar.slog(4) << "°";
  }
}

const tdouble bn_manager::calc_condProb(std::vector< rv_obs_pair* >& ilt)
{
  #if FLX_DEBUG
    if (!has_topology) throw FlxException_Crude("bn_manager::calc_condProb_1");
  #endif
  // activate information (observations)
    // reset b-tree
      const tuint S2 = treevec.size();
      for (tuint i=0;i<S2;++i) {
	treevec[i]->deactivate();
      }
    const tuint S = ilt.size();
    for (tuint i=0;i<S;++i) {
      if (ilt[i]->get_rv().get_obsrvType()!=noObsrv) {
	std::ostringstream ssV;
	ssV << "The random variable '" << ilt[i]->get_rv().name << "' has been observed.";
	throw FlxException("bn_manager::calc_condProb_2", ssV.str() );
      }
      ilt[i]->get_rv().observe(ilt[i]->get_observedStates());
      treevec[ilt[i]->get_rv().get_btreeID()-1]->set_active();
    }
    has_topology = true;	// ... because of temporary observation (did not change base-tree topology)
  // calculate probability
    tdouble up = ONE;
    tdouble lp = ONE;
    for (tuint i=0;i<S2;++i) {
      if (treevec[i]->is_active()) {
	lp*=treevec[i]->get_prob();
	up*=calc_prob(i+1);
      }
    }
  // deactivate information (observations)
    for (tuint i=0;i<S;++i) {
      ilt[i]->get_rv().unobserve();
    }
    has_topology = true; 	// ... because of temporary observation (did not change base-tree topology)
  return up/lp;
}

const flxVec bn_manager::calc_condProbVec(bn_rv_base& rv)
{
  #if FLX_DEBUG
    if (!has_topology) throw FlxException_Crude("bn_manager::calc_condProbVec_1");
  #endif
  const tuint S = rv.stateN;
    flxVec resV(S);
  if (smp_mthd==2) {
    // the random variables to determine to probability of 
      std::vector< bn_rv_base* > ilt;
      ilt.push_back(&rv);
    // run some preliminary tests
      const tuint S2 = ilt.size();	// up to now, S2 is always 1
      for (tuint i=0;i<S2;++i) {
	if (ilt[i]->get_obsrvType()!=noObsrv) {
	  std::ostringstream ssV;
	  ssV << "The random variable '" << ilt[i]->name << "' has been observed.";
	  throw FlxException("bn_manager::calc_condProbVec_2", ssV.str() );
	}
      }
    // perform the actual elimination
      bn_elim_el_table* upt = calc_condProb_Elimination(ilt);
    // perform some checks
      if (upt->get_tableSize()!=S || upt->has_rvs()!=1 || upt->depends_on(&rv)==false ) {
	delete upt;
	throw FlxException_Crude("bn_manager::calc_condProbVec_3");
      }
    // map to a vector
      tdouble *t = upt->get_table_link();
      tdouble* resVp = resV.get_tmp_vptr();
      for (tuint i=0;i<S;++i) {
	resVp[i] = t[i];
      }
    // free memory
      delete upt;
    return resV;
  }
  // activate information (observations)
    // reset b-tree
      const tuint S2 = treevec.size();
      for (tuint i=0;i<S2;++i) {
	treevec[i]->deactivate();
      }
    if (rv.get_obsrvType()!=noObsrv) {
      std::ostringstream ssV;
      ssV << "The random variable '" << rv.name << "' has been observed.";
      throw FlxException("bn_manager::calc_condProbVec_4", ssV.str() );
    }
    rv.set_locked(true);
    treevec[rv.get_btreeID()-1]->set_active();
    has_topology = true;	// ... because of temporary observation (did not change base-tree topology)
  // calculate probability
    treevec[rv.get_btreeID()-1]->clear_topo_sub();
    tdouble pg=ONE;
    // reduce base tree
      std::list<bn_rv_base*> curL;
      treevec[rv.get_btreeID()-1]->assemble_bList(curL);
      pg *= reduce_btree(curL);
      pg /= treevec[rv.get_btreeID()-1]->get_prob();
      // check if rv is still part of curL
	bool f = false;
	for (std::list<bn_rv_base*>::reverse_iterator rit=curL.rbegin(); rit!=curL.rend(); ++rit) {
	  bn_rv_base* rvt = (*rit);
	  if (rvt == &rv) {
	    f = true;
	    break;
	  }
	}
      if (f) {		// rv is still in curL
	switch (smp_mthd) {
	  case 0:
	  {
	    bn_comp* ct = extract_tree(rv.get_btreeID(), curL);
	    try {
	      // output
		GlobalVar.slog(4) << "    ... combined computation tree: ";
		ct->write(GlobalVar.slog(4));
		GlobalVar.slog(4) << std::endl;
	      for (tuint i=0;i<S;++i) {
		rv.observe(i);
		resV[i] = pg*ct->compute();
	      }
	    } catch (FlxException &e) {
	      FLXMSG("bn_manager::calc_condProbVec_5",1);
	      delete ct;
	      throw;
	    }
	    delete ct;
	  }
	    break;
	  case 1:
	    for (tuint i=0;i<S;++i) {
	      rv.observe(i);
	      const tdouble t = calc_prob_LWsim(rv.get_btreeID(),curL);
	      resV[i] = pg*t;
	    }
	    break;
	  case 2:
	  {
	    throw FlxException_Crude("bn_manager::calc_condProbVec_6");
	  }
	  default:
	    throw FlxException_Crude("bn_manager::calc_condProbVec_7");
	}
      } else {		// rv has been reduced out
	pg /= rv.get_prob_const();	// correct pg
	if (curL.size()>0) {
	  switch (smp_mthd) {
	    case 0:
	      pg *= calc_prob_sum(rv.get_btreeID(),curL);
	      break;
	    case 1:
	      pg *= calc_prob_LWsim(rv.get_btreeID(),curL);
	      break;
	    default:
	      throw FlxException_Crude("bn_manager::calc_condProbVec_8");
	  }
	}
	for (tuint i=0;i<S;++i) {
	  rv.observe(i);
	  resV[i] = pg*rv.get_prob_const();
	}
      }
  // deactivate information (observations)
    rv.set_locked(false);
    rv.unobserve();
    has_topology = true; 	// ... because of temporary observation (did not change base-tree topology)
  return resV;
}

const tdouble bn_manager::calc_probObservation() const
{
  const tuint S2 = treevec.size();
  tdouble lp = ONE;
  for (tuint i=0;i<S2;++i) {
      lp*=treevec[i]->get_prob();
  }
  return lp;
}

const tdouble bn_manager::calc_expUtility()
{
  #if FLX_DEBUG
    if (!has_topology) throw FlxException_Crude("bn_manager::calc_expUtility_1");
  #endif
  switch (smp_mthd) {
    case 0:
      return calc_expUtility_sum();
    case 1:
      return calc_expUtility_LWsim();
    case 2:
      return calc_expUtility_elim();
    default:
      throw FlxException_Crude("bn_manager::calc_expUtility_2");
  }
}

bn_elim_el_table* bn_manager::calc_condProb_Elimination(const std::vector< bn_rv_base* >& ilt)
{
  #if FLX_DEBUG
    if (!has_topology) throw FlxException_Crude("bn_manager::calc_condProb_Elimination_1");
  #endif
  // activate information (observations)
    // reset b-tree
      const tuint S2 = treevec.size();
      for (tuint i=0;i<S2;++i) {
	treevec[i]->deactivate();
      }
    const tuint S = ilt.size();
    for (tuint i=0;i<S;++i) {
      ilt[i]->set_locked(true);
      treevec[ilt[i]->get_btreeID()-1]->set_active();
      treevec[ilt[i]->get_btreeID()-1]->clear_topo_sub();
    }
  // calculate probability
    tdouble up = ONE;
    tdouble lp = ONE;
    std::list<bn_elim_el_base*> tblLst;
    for (tuint i=0;i<S2;++i) {
      if (treevec[i]->is_active()) {
	std::list<bn_rv_base*> curL;
	treevec[i]->assemble_bList(curL);
	up *= reduce_btree(curL);
	extract_tableList(tblLst,i+1,curL);
	lp*=treevec[i]->get_prob();
      }
    }
    bn_elim_el_table* upt = calc_prob_elim(tblLst);
  // check results ...
    tdouble *t = upt->get_table_link();
    const tuint tS = upt->get_tableSize();
    tdouble s = ZERO;
    lp /= up;
    for (tuint i=0;i<tS;++i) {
      s += t[i];
      t[i] /= lp;
    }
    if (fabs(s-lp)>GlobalVar.TOL()) {
      throw FlxException_Crude("bn_manager::calc_condProb_Elimination_3");
    }
  // deactivate information (locked)
    for (tuint i=0;i<S;++i) {
      ilt[i]->set_locked(false);
    }
  return upt;
}





