/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2017 Wolfgang Betz
 *
 * Fesslix is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fesslix is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fesslix.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef flxbn_flxbn_manager_H
#define flxbn_flxbn_manager_H

#include "flxMtx.h"
#include <vector>
#include <map>
#include <list>

namespace flxbn {


// -------------------------- nodes ----------------------------------------------------------------------

class bn_node {
  protected:
    bn_node* next;
    
    void check_cpt(flxVec& cpt, const tuint nrows, const tuint ncols, const tdouble roundoff_correct) const;
  public:
    bn_node() : next(NULL) {}
    virtual ~bn_node();
    
    void set_node(bn_node* n2s) { if (next) next->set_node(n2s); else next = n2s; }
};


class bn_node_first : public bn_node {
  public:
};



// -------------------------- random variable ----------------------------------------------------------------------

enum bn_rv_type { trueRV, detRel, decNode };
enum bn_observation { noObsrv, snglObsrv, multObsrv };

class bn_rv_base;

class rv_edge {
  private:
    int btreeID;	// basis tree (just single observations)
    int streeID;	// sub-tree of basis tree
    /**
    * @brief clears the topology information
    */
    void clear_topo(const bool full, const bn_rv_base* const owner);
  public:
    bn_rv_base* const rv;
    
    rv_edge(bn_rv_base* const rv) : btreeID(0),streeID(0), rv(rv) {}
    rv_edge& operator=(const rv_edge& rhs);
    
    /**
    * @return -1 if the link is observed (both ends are observed) .. otherwise the tree-ID (0 if not assigned)
    */
    const int get_btreeID() const { return btreeID; }
    const int get_streeID() const { return streeID; }
    
    friend class bn_rv_base;
};

class rv_obs_pair {
  private:
    bn_rv_base* const rv;
    std::vector<tuint> observedStates;
  public:
    rv_obs_pair(const rv_obs_pair& rhs) : rv(rhs.rv), observedStates(rhs.observedStates) {}
    rv_obs_pair(bn_rv_base* const rv, const std::vector<tuint> observedStates) : rv(rv), observedStates(observedStates) {}
    void observe() const;
    bn_rv_base& get_rv() const { return *rv; }
    const std::vector<tuint> get_observedStates() const { return observedStates; }
};

class bn_manager;

/**
* @brief the base class
*/
class bn_rv_base {
  private:
    tuint ID;		// starts counting with 0
    int btreeID;		// starts counting with 1; 0: not dealt with; -1: fixed (not directly saved)
    int streeID;		// ""; -2: removed
    const tuint parentN;
    std::vector<rv_edge> parents;
    tuint childN;
    std::vector<rv_edge> childs;
    bn_observation obsrvType;
    std::vector<tuint> observedStates;
    bool locked;
    
    /**
    * @brief set crv as a new child of this variable
    */
    void set_child(bn_rv_base* crv);
  protected:
    /**
    * @brief get the link of the child 'crv'
    */
    rv_edge& get_child(bn_rv_base* crv);
    /**
    * @brief get the link of the parent 'prv'
    */
    rv_edge& get_parent(bn_rv_base* prv);
  public:
    /**
    * @brief name of the random variable
    */
    const std::string name;
    /**
    * @brief true, if it can be used to connect other RVs
    */
    const bool outputRV;
    /**
    * @brief number of possible states
    */
    const tuint stateN;
    const std::vector<std::string>& states;
    /**
    * @brief object containing the RV
    */
    bn_node* const node;
    /**
    * @brief list of the current states, the RVs are in
    */
    static tuint* stateIDs;
    
    bn_rv_base(const std::string name, const bool outputRV, const std::vector<std::string>& states, const std::vector<bn_rv_base*> drvs, bn_node* const node);
    virtual ~bn_rv_base() {}
    /**
    * @brief returns the type
    * 	1: true random variable
    * 	2: deterministic relation
    */
    virtual const bn_rv_type get_type() const = 0;
    /**
    * @brief set a state of the variable randomly - consider observations and parent-states
    */
    virtual const bool rnd_state(tdouble& p) const = 0;
    virtual void rnd_state(tdouble& p,tdouble& p2) const = 0;
    /**
    * @brief conditional probability of the current configuration
    */
    virtual const tdouble get_prob_const() const = 0;
    
    const tuint get_ID() const { return ID; }
    const int get_btreeID() const { return btreeID; }
    const int get_streeID() const { return streeID; }
    const tuint get_parentN() const { return parentN; }
    const tuint get_childN() const { return childN; }
    const rv_edge& get_parent(const tuint pid) const { return parents[pid]; }
    const rv_edge& get_child(const tuint cid) const { return childs[cid]; }
    const bn_observation get_obsrvType() const { return obsrvType; }
    const tuint get_observedStates(const tuint i) const { return observedStates[i]; }
    const tuint get_observedStatesN() const { return observedStates.size(); }
    const bool is_state_observed(const tuint stateID) const;
    const bool is_locked() const { return locked; }
    void set_locked(const bool lv) { locked = lv; }
    /**
    * @brief returns the ID of 'ts' ... starts counting with 0
    * @param ts lowercase string (state)
    */
    const tuint get_stateID(const std::string ts) const;
    /**
    * @brief returns a new streeID based on observation & co.
    */
    const int get_streeID_new() const;
    
    void set_ID(const tuint IDv);
    void observe(const tuint sid);
    virtual void observe(const std::vector<tuint> oV);
    virtual void unobserve();
    /**
    * @brief clears the topology information
    * @param full true: clear topo of base and sub-tree; false: clear topo of sub-tree only
    */
    void clear_topo(const bool full=true);
    /**
    * @brief remove the RV from the corresponding sub-tree
    */
    void stree_remove();
    /**
    * @brief detects the topology of the tree
    * @param tid id of the current tree to detect
    * @param tdir travel direction (0:start; 1:down; 2:up)
    * @param last last rv (so that we do not go back, acidentally)
    */
    void detect_topo(const int tid, const int tdir, bn_rv_base* last);
    /**
    * @brief detects the topology of the sub-tree
    * @param btid id of the current base-tree to detect
    * @param stid id of the current sub-tree to detect
    * @param tdir travel direction (0:start; 1:down; 2:up)
    * @param last last rv (so that we do not go back, acidentally)
    */
    void detect_topo_sub(const int btid, const int stid, const int tdir, bn_rv_base* last);
    /**
    * @brief distance to drv (only for parents and childs AND siblings)
    * @return infinity, if not parent or child; 0: if drv==this or fix; 1: otherwise
    */
    const tuint get_dist(const bn_rv_base* const drv) const;
    /**
    * @brief vist[i] is true, if variable is in the sub-tree
    */
    void get_subtree_rvs(bool *vist,const int streeIDv) const;
};

/**
* @brief a BN random variable
*/
class bn_rv : public bn_rv_base {
  private:
    
    /**
    * @brief number of columns in the conditioned probability table
    */
    tuint tableCols;
    /**
    * @brief stores the indices to compute a column
    */
    iVec tableColsVec;
    /**
    * @brief determines the non-zero columns given the parent-states
    */
    iVec tableCol0;
    /**
    * @brief determines the probability-sum of all non-zero columns (for multObserv)
    */
    flxVec tableCol0sum;
    /**
    * @brief conditional probability table
    */
    const flxVec& table;
    
    /**
    * @brief returns the current column - based on the parent states
    */
    const tuint get_curr_col() const;
    /**
    * @brief updates tableCol0 and tableCol0sum
    */
    void update_tableCol0();
    
    virtual void observe(const std::vector<tuint> oV);
    virtual void unobserve();
  public:
    bn_rv(const std::string name, const bool outputRV, const std::vector<std::string>& states, const std::vector<bn_rv_base*> drvs, const flxVec& tableV, bn_node* const node);
    virtual const bn_rv_type get_type() const { return trueRV; }
    virtual const bool rnd_state(tdouble& p) const;
    virtual void rnd_state(tdouble& p,tdouble& p2) const;
    virtual const tdouble get_prob_const() const { return table[tableCols*stateIDs[get_ID()]+get_curr_col()]; }
    const tdouble *const get_table_ptr() const { return table.get_tmp_vptr_const(); }
};


/**
* @brief BN deterministic relationship (base class)
*/
class bn_detRel : public bn_rv_base {
  public:
    bn_detRel(const std::string name, const bool outputRV, const std::vector<std::string>& states, const std::vector<bn_rv_base*> drvs, bn_node* const node);
    virtual const bn_rv_type get_type() const { return detRel; }
    
    const bool depends_on(bn_rv_base* rv);
    virtual void sum_out(const tuint Nas, const tuint* const IDas, bn_rv_base* rv, tdoublePtr& resVec) = 0;
};


/**
* @brief BN deterministic relationship (and)
*/
class bn_detRel_and : public bn_detRel {
  private:
    const bool eval() const;
  public:
    bn_detRel_and(const std::string name, const bool outputRV, const std::vector<std::string>& states, const std::vector<bn_rv_base*> drvs, bn_node* const node);
    virtual const bool rnd_state(tdouble& p) const;
    virtual void rnd_state(tdouble& p,tdouble& p2) const;
    virtual const tdouble get_prob_const() const { return eval()?ONE:ZERO; }
    virtual void sum_out(const tuint Nas, const tuint* const IDas, bn_rv_base* rv, tdoublePtr& resVec);
};


/**
* @brief BN decission
*/
class bn_dec : public bn_rv_base {
  private:
    virtual void observe(const std::vector<tuint> oV);
  public:
    bn_dec(const std::string name, const bool outputRV, const std::vector<std::string>& states, bn_node* const node);
    virtual const bn_rv_type get_type() const { return decNode; }
    virtual const bool rnd_state(tdouble& p) const;
    virtual void rnd_state(tdouble& p,tdouble& p2) const;
    virtual const tdouble get_prob_const() const;
    /**
    * @brief check if decission has been observed
    */
    void check_valid() const;
};


// -------------------------- Utility ----------------------------------------------------------------------

class bn_comp;
class Btree_class;
typedef bn_comp* bn_comp_ptr;

class bn_utility {
  private:
    const std::string name;
    const tuint parentN;
    std::vector<rv_edge> parents;
    bn_node* const node;
    /**
    * @brief number of columns in the conditioned probability table
    */
    tuint tableCols;
    /**
    * @brief stores the indices to compute a column
    */
    iVec tableColsVec;
    /**
    * @brief utility vector
    */
    const flxVec& utilityVec;
    
    /**
    * @brief returns the current column - based on the parent states
    */
    const tuint get_curr_col() const;
    
  public:
    
    bn_utility(const std::string name, const std::vector< bn_rv_base* > drvs, const flxVec& utilityVec, bn_node*const node);
//     const bool rnd_state(tdouble& p) const;
//     void rnd_state(tdouble& p,tdouble& p2) const;
    /**
    * @brief returns the current utility
    */
    const tdouble get_utility_const() const { return utilityVec[get_curr_col()]; }
    const std::string& get_name() const { return name; }
    
    /**
    * @brief observe all parents which are not snglObsrv
    * @param lock true: lock; false: unlock
    */
    bn_comp* lock_parents(std::vector<Btree_class*> &treevec, bn_comp_ptr& ct_c);
    void unlock_parents();
    /**
    * @brief activates the base-trees of the parents (and lock the parents)
    */
    void activate_tree(std::vector<Btree_class*> &treevec) const;
    /**
    * @brief returns the parents of this utility
    * @param PVec list with parents ... must be empty in the beginning
    */
    void get_parents(std::vector< bn_rv_base* >& pVec) const;
};


// -------------------------- Computation tree ----------------------------------------------------------------------

class bn_comp {
  private:
    /**
    * @brief 0:no sum; 1:full sum; 2:not-full sum; 3:det-relation; 4: just multiplicator
    */
    const tuint sum_id;
    /**
    * @brief the RV of the object (either as sum or as value)
    */
    bn_rv_base* const rv;
    bn_utility* const ut;
    std::vector<bn_comp*> childs;
    
    bn_comp* find_sum(bn_rv_base* const rvs);
  public:
    bn_comp(const bool sumb,bn_rv_base* const rv);
    bn_comp(bn_utility* const ut);
    
    ~bn_comp();
    /**
    * @brief append a sum
    */
    bn_comp* append(bn_rv_base* const rvs);
    /**
    * @brief append a variable (not a sum)
    * @param rv the RV to append
    * @param chk_id 0: if extern call; 
    */
    void append(bn_rv_base* const rvs, const tuint chk_id);
    /**
    * @brief append a computation tree
    */
    void append(bn_comp* ct);
    /**
    * @brief optimize the summation tree
    */
    const tdouble compute() const;
    void write(std::ostream& sout) const;
};

// -------------------------- elimination element class ----------------------------------------------------------------------

class bn_elim_el_base {
  protected:
    
  public:
    virtual ~bn_elim_el_base() {}
    /**
    * @brief returns true if the element depends on the given random variable rv
    */
    virtual const bool depends_on(bn_rv_base* rv) = 0;
    /**
    * @brief sum out rv; resVec is multiplied with the probabilities of the individual states
    * @param rv to sum out
    * @param Nas number of active states
    * @param IDas IDs of active states
    * @param resVec the result (size: Nas)
    */
    virtual void sum_out(const tuint Nas, const tuint* const IDas, bn_rv_base* rv, tdoublePtr& resVec) = 0;
    /**
    * @brief return coefficient of the current configuration
    */
    virtual const tdouble get_prob() const = 0;
    /**
    * @brief return the variables of the 'element' and appends them to arvlist
    */
    virtual void get_active_RVs(std::vector<bn_rv_base*>& arvlist) = 0;
};

class bn_elim_el_table_base : public bn_elim_el_base {
  protected:
    tuint Nvars;		// number of variables in the list
    bn_rv_base** vars;		// variables
    tuint* Nstates;		// number of states of each variable
    tuint* mapTable;		// mapping table for each variable (positioning rool within table)
    tuint** curID;		// current IDs of the variables  
    const tdouble get_prob_v(const tdouble* const table) const;
  public:
    /**
    * @brief create it from a random variabe
    */
    bn_elim_el_table_base(bn_rv* rv);
    bn_elim_el_table_base();
    virtual ~bn_elim_el_table_base();
    
    /**
    * @brief returns true if the list contains the random variable 'rv'
    */
    const tuint has_rvs() const { return Nvars; }
    const bool depends_on(bn_rv_base* rv);
    void get_active_RVs(std::vector<bn_rv_base*>& arvlist);
};

class bn_elim_el_table_rv : public bn_elim_el_table_base {
  protected:
    const tdouble* const table;		// finally, the table :)
  public:
    bn_elim_el_table_rv(bn_rv* rv);
    
    void sum_out(const tuint Nas, const tuint* const IDas, bn_rv_base* rv, tdoublePtr& resVec);
    const tdouble get_prob() const;
};

class bn_elim_el_table : public bn_elim_el_table_base {
  protected:
    tuint tableSize;		// the size of the table
    tdouble* table;		// finally, the table :)
    bool deleteTable;		// does the memory of the table have to be deleted?
  public:
    /**
    * @brief create a new table by summing out 'rv' from subLst ... the tables in subLst will be deleted
    */
    bn_elim_el_table(bn_rv_base* rv, std::list< bn_elim_el_base* >& subLst);
    /**
    * @brief create a new table by combining the tables in subLst ... the tables in subLst will be deleted
    */
    bn_elim_el_table(std::list< bn_elim_el_base* >& subLst);
    bn_elim_el_table(const tdouble p);
    ~bn_elim_el_table();
    void sum_out(const tuint Nas, const tuint* const IDas, bn_rv_base* rv, tdoublePtr& resVec);
    const tdouble get_prob() const;
    tdouble* get_table_link() { return table; }
    const tuint get_tableSize() { return tableSize; }
    
    /**
    * @brief compute the expected utility of utl w.r.t. the probabilities in this table
    */
    const tdouble exp_Utility_of_table(const bn_utility& utl) const;
};

class bn_elim_el_detRel : public bn_elim_el_base {
  protected:
    bn_detRel* detRel;
  public:
    bn_elim_el_detRel(bn_detRel* detRel) : detRel(detRel) {}
    ~bn_elim_el_detRel() {}
    
    const bool depends_on(bn_rv_base* rv) { return detRel->depends_on(rv); }
    void sum_out(const tuint Nas, const tuint* const IDas, bn_rv_base* rv, tdoublePtr& resVec) { detRel->sum_out(Nas,IDas,rv,resVec); }
    const tdouble get_prob() const;
    void get_active_RVs(std::vector<bn_rv_base*>& arvlist);
};

// -------------------------- Manager ----------------------------------------------------------------------

class Btree_class {
  private:
    /**
    * @brief true, if the base-tree is currently active
    */
    bool active;
    /**
    * @brief probability value of the tree
    */
    tdouble prob;
    /**
    * @brief lists all RV's in the base tree
    */
    std::vector<bn_rv_base*> btrvs;
  public:
    const int ID;
    /**
    * @brief root-node of the tree
    */
    bn_rv_base* const root;
    
    Btree_class(const int ID, bn_rv_base* const root, const std::vector<bn_rv_base*>& rvid);
    Btree_class& operator=(const Btree_class& rhs);
    
    void deactivate() { active=false; }
    void set_active() { active=true; }
    const bool is_active() const { return active; }
    const tdouble get_prob() const { return prob; }
    void set_prob();
    void clear_topo_sub();
    void assemble_bList(std::list<bn_rv_base*>& curL) const;
};

class bn_manager {
  private:
    /**
    * @brief true: if topology has been set up
    */
    bool has_topology;
    /**
    * @brief lists all existing RVs
    */
    std::map<std::string, bn_rv_base*> rvs; 
    std::vector<bn_rv_base*> rvid;
    /**
    * @brief list containing all utilities
    */
    std::vector<bn_utility*> utvec;
    /**
    * @brief (basis) tree - topology
    */
    std::vector<Btree_class*> treevec; 	// tree_index = vector_index + 1
    /**
    * @brief elimination order for variable elimination (starting with first, ending with last)
    */
    std::list<bn_rv_base*> elimOrder;
    /**
    * @brief link to the list of nodes
    */
    bn_node_first nodes;
    /**
    * @brief the method to obtain the probability
    * 0: recursive summation (exact propagation)
    * 1: Likelihood Weighting (stochastic simulation)
    * 2: variable elimination (exact propagation)
    */
    tuint smp_mthd;
    tuint smp_mthd_2_N;
    bool smp_mthd_2_eP;
    
    /**
    * @brief reduces a given tree
    * @param curL a list with all variables -> all variables in reduced tree
    * @return probability value which could be condensed out
    * @note it is important to call clear_topo_sub first !!!
    */
    const tdouble reduce_btree(std::list<bn_rv_base*>& curL);
    /**
    * @brief extracts the computation tree from the reduced base tree
    */
    bn_comp* extract_tree(const int btreeID, std::list<bn_rv_base*>& curL);
      void correct_vistVec(bool *vist, std::list<bn_rv_base*>& curL) const;
    /**
    * @brief extracts the table list (for variable elimination) from the reduced base tree
    */
    void extract_tableList(std::list<bn_elim_el_base*>& tblLst, const int btreeID, std::list<bn_rv_base*>& curL);
    /**
    * @brief computes the probability ( !!! on the base-tree !!! )
    */
    const tdouble calc_prob(const int btreeID);
      /**
      * @brief computes the probability exactly using summation ( !!! on sub-trees of the REDUCED base-tree !!! )
      */
      const tdouble calc_prob_sum(const int btreeID, std::list<bn_rv_base*>& curL);
      /**
      * @brief computes the probability exactly using variable elimination ( !!! on sub-trees of the REDUCED base-tree !!! )
      */
      const tdouble calc_prob_elim(const int btreeID, std::list<bn_rv_base*>& curL);
      /**
      * @brief computes the probability exactly using summation
      * @param tblLst list with tables (probabilities) --> will contain the returned probabilities
      * @param return probabilities (pure)
      */
      bn_elim_el_table* calc_prob_elim(std::list<bn_elim_el_base*>& tblLst);
      /**
      * @brief computes the probability using MCI ( !!! on the complete REDUCED base-tree !!! )
      */
      const tdouble calc_prob_LWsim(const int btreeID, std::list<bn_rv_base*>& curL);
      /**
      * @brief calc. expected utility using summation
      */
      const tdouble calc_expUtility_sum();
      /**
      * @brief calc. expected utility using elimination
      */
      const tdouble calc_expUtility_elim();
      /**
      * @brief calc. expected utility using MCI
      */
      const tdouble calc_expUtility_LWsim();
    /**
    * @brief updates the observed state values
    */
    void update_observedStateValues();
    /**
    * @brief computes the shortest path between random variables
    */
    void compute_shortestPath(const tuint strvsS,tuint* distRV,tuint* pathRV,bool* strv_active,std::vector<bn_rv_base*> strvs);
    /**
    * @brief checks if the shortest path between RVs is valid
    * @return true, everything went fine
    */
    const bool check_shortestPath(const tuint strvsS,tuint* distRV,tuint* pathRV,bool* strv_active,std::vector<bn_rv_base*> strvs, const tuint start, const tuint end) const;
    /**
    * @brief frees all memory in treevec
    */
    void treevec_delete();
    
  // output
    /**
    * @brief output list
    */
    void output_list(std::ostream& sout, const std::list<bn_rv_base*>& curL) const;
    
  public:
    bn_manager() : has_topology(false), smp_mthd(2), smp_mthd_2_N(tuint(1e5)), smp_mthd_2_eP(false) {}
    ~bn_manager();
    
  // configuration
    /**
    * @brief adds a node to the node-list
    */
    void set_node(bn_node* n2s) { nodes.set_node(n2s); }
    void set_smp_mthd(const tuint smp_mthdV);
      void set_smp_mthd_2_N(const tuint smp_mthd_2_NV) { smp_mthd_2_N = smp_mthd_2_NV; }
      void set_smp_mthd_2_eP(const bool smp_mthd_2_ePV) { smp_mthd_2_eP = smp_mthd_2_ePV; }
    /**
    * @brief adds a random variable to the RV-list
    */
    void add_rv(bn_rv_base* rv);
    /**
    * @brief adds a utility to the utility-list
    */
    void add_utility(bn_utility* ut);
    /**
    * @brief returns a random variable -> or throws an error
    */
    bn_rv_base& get_rv(const std::string& rvn) const;
    /**
    * @brief resets the topology (fast)
    */
    void topo_reset() { has_topology = false; }
    /**
    * @brief create the topology
    */
    void create_topo();
    /**
    * @brief observe a random variable ... conserve topology if possible
    * @param rv the random variable to observe
    * @param oV the observation
    */
    void observe_RV(bn_rv_base* rv, const std::vector<tuint> oV);
    /**
    * @brief append a variable to the elimination order
    */
    void append_elimOrderVar(bn_rv_base* rv);
    
  // compute
    /**
    * @brief calc. conditional probability
    */
    const tdouble calc_condProb(std::vector<rv_obs_pair*>& ilt);
    /**
    * @brief calc. conditional probability (combine all possible states)
    * @param ile vector with names of RVs to combine
    */
    const flxVec calc_condProbVec(bn_rv_base& rv);
    /**
    * @brief calc. probability of observed configuration
    */
    const tdouble calc_probObservation() const;
    /**
    * @brief calc. expected utility
    */
    const tdouble calc_expUtility();
    /**
    * @brief calc. conditional probability (using variable elimination) 
    */
    bn_elim_el_table* calc_condProb_Elimination(const std::vector<bn_rv_base*>& ilt);
    
  // output
    /**
    * @brief output the topology (list)
    */
    void output_topo(std::ostream& sout) const;
    /**
    * @brief output the observations
    */
    void output_observed(std::ostream& sout) const;
    
    friend class Btree_class;
};

#ifdef flxbn_flxbn_manager_CPP
  bn_manager bnm;
#else
  extern bn_manager bnm;
#endif
  


} // end of namespace 'flxbn_flxbn_manager_H'


#endif // flxbn_flxbn_manager_H

