/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2017 Wolfgang Betz
 *
 * Fesslix is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fesslix is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fesslix.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include "flxbn_obj.h"

#ifdef __WINDOWS__
  #pragma comment(lib, "flxlib")
#endif

// using namespace std;
using namespace flxbn;



bn_obj_irv::bn_obj_irv(const std::string name, const std::vector< std::string >& statesV, const flxVec& tableV, const tdouble roundoff_correct)
: states(statesV), table(tableV), rv(name,true,states,std::vector<bn_rv_base*>(),table,this)
{
  // check cpt
    check_cpt(table,states.size(),1,roundoff_correct);
  bnm.add_rv(&rv);
}

bn_obj_drv::bn_obj_drv(const std::string name, const std::vector< std::string >& statesV, const std::vector< bn_rv_base* >& drvs, const flxVec& tableV, const tdouble roundoff_correct)
: states(statesV), table(tableV), rv(name,true,states,drvs,table,this)
{
  // check cpt
    tuint tableCols = 1;
    for (tuint i=0;i<drvs.size();++i) {
      tableCols*=drvs[i]->stateN;
    }
    check_cpt(table,states.size(),tableCols,roundoff_correct);
  bnm.add_rv(&rv);
}

bn_obj_detrel_and::bn_obj_detrel_and(const std::string name, const std::vector< std::string >& statesV, const std::vector< bn_rv_base* >& drvs)
: states(statesV), drv(name,true,states,drvs,this)
{
  bnm.add_rv(&drv);
}

bn_obj_dec::bn_obj_dec(const std::string name, const std::vector< std::string >& statesV)
: states(statesV), dec(name,true,states,this)
{
  bnm.add_rv(&dec);
}

bn_obj_utility::bn_obj_utility(const std::string name, const std::vector< bn_rv_base* >& drvs, const flxVec& utilityVecV)
: utilityVec(utilityVecV), ut(name, drvs, utilityVec,this)
{
  bnm.add_utility(&ut);
}

bn_obj_train_base::bn_obj_train_base(const std::string& name, const flxVec& ivals, const flxVec& sivals, const tuint Nstages)
: name(name), NstatesR(ivals.get_N()+1), NstatesS(sivals.get_N()+2), ivalsR(ivals), ivalsS(sivals),
 cptable1S(NstatesR*NstatesS), cptable2S(NstatesS*NstatesR*NstatesS),
rV(Nstages), aV(Nstages), sV(Nstages), Nstages(Nstages) 
{
  // check plausibility
    {
      tdouble prev = ZERO;
      for (tuint i=1;i<NstatesR;++i) {
	if (prev > ivalsR[i-1]) {
	  std::ostringstream ssV;
	  ssV << "Error in rain intervals";
	  throw FlxException("bn_obj_train_base::bn_obj_train_base_1", ssV.str() );
	}
	prev = ivalsR[i-1];
      }
    }
    {
      tdouble prev = ZERO;
      for (tuint i=1;i<NstatesS-1;++i) {
	if (prev > ivalsS[i-1]) {
	  std::ostringstream ssV;
	  ssV << "Error in sum intervals";
	  throw FlxException("bn_obj_train_base::bn_obj_train_base_2", ssV.str() );
	}
	prev = ivalsS[i-1];
      }
    }
  // assign states R & S
    for (tuint i=0;i<NstatesR;++i) {
      std::ostringstream ssV;
      ssV << i;
      statesR.push_back(ssV.str());
    }
    for (tuint i=0;i<NstatesS;++i) {
      std::ostringstream ssV;
      ssV << i;
      statesS.push_back(ssV.str());
    }
    // compute cptable1S
    cptable1S[0] = ONE;
    tdouble r_last = ZERO;
    tdouble r_next;
	#ifndef FLX_CV_1
		tdouble pa[NstatesS-1];
	#else
		tdouble *pa = new tdouble[NstatesS-1];
	#endif
    for (tuint i=1;i<NstatesR;++i) {
      r_next = ivalsR[i-1];
      uniAdd(ZERO,ZERO,r_last,r_next,pa);
      for (tuint j=1;j<NstatesS;++j) {
	cptable1S[j*NstatesR+i] = pa[j-1];
      }
      r_last = r_next;
    }

  // compute cptable2S
    // it hasn't rained so far
      for (tuint j=0;j<NstatesS;++j) {
	for (tuint i=0;i<NstatesR;++i) {
	  cptable2S[j*NstatesR*NstatesS+i] = cptable1S[j*NstatesR+i];
	}
      }
    // all finite intervals of rain
      tdouble s_last = ZERO;
      tdouble s_next;
      for (tuint j=1;j<NstatesS-1;++j) {
	s_next = ivalsS[j-1];
	// it hasn't rained in the last step
	  cptable2S[j*NstatesR*NstatesS+j*NstatesR] = ONE;
	// there was rain in the last step
	  r_last = ZERO;
	  for (tuint i=1;i<NstatesR;++i) {
	    r_next = ivalsR[i-1];
	    uniAdd(s_last,s_next,r_last,r_next,pa);
	    for (tuint k=1;k<NstatesS;++k) {
	      cptable2S[k*NstatesR*NstatesS+j*NstatesR+i] = pa[k-1];
	    }
	    r_last = r_next;
	  }
	s_last = s_next;
      }
    // the finite column
      for (tuint i=0;i<NstatesR;++i) {
	cptable2S[(NstatesS-1)*NstatesR*NstatesS+(NstatesS-1)*NstatesR+i] = ONE;
      }
	#ifdef FLX_CV_1
		delete [] pa;
	#endif
}

bn_obj_train_base::~bn_obj_train_base()
{
  free_mem();
}

void bn_obj_train_base::uniAdd(tdouble a1, tdouble b1, tdouble a2, tdouble b2, tdouble res[]) const
{
  for (tuint i=0;i<NstatesS-1;++i) res[i] = ZERO;
  tdouble d1 = b1 - a1;
  tdouble d2 = b2 - a2;
  if (d1>d2) {		// d2 has to be the larger interval
    tdouble t = d2; d2 = d1; d1 = t;
    t = b2; b2 = b1; b1 = t;
    t = a2; a2 = a1; a1 = t;
  }
  #if FLX_DEBUG
    if (d1<ZERO||d2<ZERO) throw FlxException_Crude("bn_obj_train::uniAdd_1");
  #endif
  const tdouble m = (a1+a2+b1+b2)/2;
  if (d2<=GlobalVar.TOL()) {			// deterministic addition
    for (tuint i=1;i<NstatesS-1;++i) {
      if (ivalsS[i-1]>=m) {
	res[i-1] = ONE;
	return;
      }
    }
    res[NstatesS-2] = ONE;
    return;
  }
  const tdouble h = ONE/d2;
  const tdouble am = (d2-d1)*h;
  const tdouble as = d1/2*h;
  const tdouble x1 = m-d2/2-d1/2;
  const tdouble x2 = m-d2/2+d1/2;
  const tdouble x3 = m+d2/2-d1/2;
  const tdouble x4 = m+d2/2+d1/2;
  tdouble prev_iv = ZERO;
  tdouble next_iv;
  tdouble As = ZERO;
  for (tuint i=1;i<NstatesS-1&&prev_iv<x4;++i) {
    next_iv = ivalsS[i-1];
    tdouble A = ZERO;
    if (next_iv>x1 && prev_iv<x2) {		// section 1
      const tdouble t_s = (prev_iv<=x1)?ZERO:(prev_iv-x1)/(x2-x1);
      const tdouble t_e = (next_iv>=x2)?ONE:(next_iv-x1)/(x2-x1);
      A += (t_e*t_e-t_s*t_s)*as;
    }
    if (next_iv>x2 && prev_iv<x3) {		// section 2
      const tdouble t_s = (prev_iv<=x2)?ZERO:(prev_iv-x2)/(x3-x2);
      const tdouble t_e = (next_iv>=x3)?ONE:(next_iv-x2)/(x3-x2);
      A += (t_e-t_s)*am;
    }
    if (next_iv>x3 && prev_iv<x4) {		// section 3
      const tdouble t_s = (prev_iv<=x3)?ZERO:(prev_iv-x3)/(x4-x3);
      const tdouble t_e = (next_iv>=x4)?ONE:(next_iv-x3)/(x4-x3);
      A += ((ONE-t_s)*(ONE-t_s)-(ONE-t_e)*(ONE-t_e))*as;
    }
    res[i-1] = A;
    As += A;
    prev_iv = next_iv;
  }
  tdouble t = ONE - As;
  if (fabs(t)<=GlobalVar.TOL() && t < ZERO) t = ZERO;
  #if FLX_DEBUG
    if (t<ZERO) {
      throw FlxException_Crude("bn_obj_train::uniAdd_2");
    }
  #endif
  if (x4>prev_iv) {
    res[NstatesS-2] = t;
  }
}

void bn_obj_train_base::free_mem()
{
  for (tuint i=0;i<Nstages;++i) {
    if (rV[i]) {
      delete rV[i];
      rV[i] = NULL;
    }
    if (aV[i]) {
      delete aV[i];
      aV[i] = NULL;
    }
    if (sV[i]) {
      delete sV[i];
      sV[i] = NULL;
    }
  }
}

bn_obj_train::bn_obj_train(const std::string& name, const flxVec& ivals, const flxVec& ptable, const flxVec& cptable, const flxVec& rcptable, const flxVec& sivals, const tuint Nstages, const tdouble roundoff_correct)
: bn_obj_train_base(name,ivals,sivals,Nstages),
ptableR(ptable), cptableR(cptable), cptableA(rcptable)
{
  // check plausibility
    check_cpt(ptableR,NstatesR,1,roundoff_correct);
    check_cpt(cptableR,NstatesR,NstatesR,roundoff_correct);
    check_cpt(cptableA,NstatesR,NstatesR,roundoff_correct);
  // create BN-structure
  try {
    // rain
      {
	std::ostringstream ssV;
	ssV << name << "_r0";
	rV[0] = new bn_rv(ssV.str(),false,statesR,std::vector<bn_rv_base*>(),ptableR,this);
	bnm.add_rv(rV[0]);
      }
      for (tuint i=1;i<Nstages;++i) {
	std::vector<bn_rv_base*> dv;
	dv.push_back(rV[i-1]);
	std::ostringstream ssV;
	ssV << name << "_r" << i;
	rV[i] = new bn_rv(ssV.str(),false,statesR,dv,cptableR,this);
	bnm.add_rv(rV[i]);
      }
    // sum
      {
	std::vector<bn_rv_base*> dv;
	dv.push_back(rV[0]);
	std::ostringstream ssV;
	ssV << name << "_s0";
	sV[0] = new bn_rv(ssV.str(),false,statesS,dv,cptable1S,this);
	bnm.add_rv(sV[0]);
      }
      for (tuint i=1;i<Nstages;++i) {
	std::vector<bn_rv_base*> dv;
	dv.push_back(sV[i-1]);
	dv.push_back(rV[i]);
	std::ostringstream ssV;
	ssV << name << "_s" << i;
	sV[i] = new bn_rv(ssV.str(),false,statesS,dv,cptable2S,this);
	bnm.add_rv(sV[i]);
      }
    // radar
      for (tuint i=0;i<Nstages;++i) {
	std::vector<bn_rv_base*> dv;
	dv.push_back(rV[i]);
	std::ostringstream ssV;
	ssV << name << "_a" << i;
	aV[i] = new bn_rv(ssV.str(),false,statesR,dv,cptableA,this);
	bnm.add_rv(aV[i]);
      }
  } catch (FlxException &e) {
    FLXMSG("bn_obj_train::bn_obj_train",1);
    free_mem();
    throw;
  }
}

bn_obj_train_radar::bn_obj_train_radar(const std::string& name, const flxVec& ivals, const flxVec& cptable_RonA, const flxVec& sivals, const tuint Nstages, const tdouble roundoff_correct)
: bn_obj_train_base(name, ivals, sivals, Nstages), cptableRonA(cptable_RonA)
{
  // check plausibility
    check_cpt(cptableRonA,NstatesR,NstatesR,roundoff_correct);
  // create BN-structure
  try {
    // radar observations
      for (tuint i=0;i<Nstages;++i) {
	std::ostringstream ssV;
	ssV << name << "_a" << i;
	aV[i] = new bn_dec(ssV.str(),false,statesR,this);
	bnm.add_rv(aV[i]);
      }
    // rain
      for (tuint i=0;i<Nstages;++i) {
	std::vector<bn_rv_base*> dv;
	dv.push_back(aV[i]);
	std::ostringstream ssV;
	ssV << name << "_r" << i;
	rV[i] = new bn_rv(ssV.str(),false,statesR,dv,cptableRonA,this);
	bnm.add_rv(rV[i]);
      }
    // sum
      {
	std::vector<bn_rv_base*> dv;
	dv.push_back(rV[0]);
	std::ostringstream ssV;
	ssV << name << "_s0";
	sV[0] = new bn_rv(ssV.str(),false,statesS,dv,cptable1S,this);
	bnm.add_rv(sV[0]);
      }
      for (tuint i=1;i<Nstages;++i) {
	std::vector<bn_rv_base*> dv;
	dv.push_back(sV[i-1]);
	dv.push_back(rV[i]);
	std::ostringstream ssV;
	ssV << name << "_s" << i;
	sV[i] = new bn_rv(ssV.str(),false,statesS,dv,cptable2S,this);
	bnm.add_rv(sV[i]);
      }
  } catch (FlxException &e) {
    FLXMSG("bn_obj_train_radar::bn_obj_train_radar",1);
    free_mem();
    throw;
  }
}



